# -*- coding: utf-8 -*-

"""The mcbkg_ps analysis is a multi-dataset time-integrated single source
analysis with a two-component likelihood function using a spacial and an energy
event PDF. It initializes the background energy pdf using auxiliary fluxes and
pdfs, which are generated by running `scripts/mceq_atm_bkg.py` script.
"""

import argparse
import logging
import numpy as np
import pickle

from skyllh.core.progressbar import ProgressBar

# Classes to define the source hypothesis.
from skyllh.physics.source import PointLikeSource
from skyllh.physics.flux import PowerLawFlux
from skyllh.core.source_hypo_group import SourceHypoGroup
from skyllh.core.source_hypothesis import SourceHypoGroupManager

# Classes to define the fit parameters.
from skyllh.core.parameters import (
    SingleSourceFitParameterMapper,
    FitParameter
)

# Classes for the minimizer.
from skyllh.core.minimizer import Minimizer, LBFGSMinimizerImpl

# Classes for utility functionality.
from skyllh.core.config import CFG
from skyllh.core.random import RandomStateService
from skyllh.core.optimize import SpatialBoxEventSelectionMethod
from skyllh.core.smoothing import BlockSmoothingFilter
from skyllh.core.timing import TimeLord
from skyllh.core.trialdata import TrialDataManager

# Classes for defining the analysis.
from skyllh.core.test_statistic import TestStatisticWilks
from skyllh.core.analysis import (
    TimeIntegratedMultiDatasetSingleSourceAnalysis as Analysis
)

# Classes to define the background generation.
from skyllh.core.scrambling import DataScrambler, UniformRAScramblingMethod
from skyllh.i3.background_generation import FixedScrambledExpDataI3BkgGenMethod

# Classes to define the signal and background PDFs.
from skyllh.core.signalpdf import RayleighPSFPointSourceSignalSpatialPDF
from skyllh.i3.signalpdf import SignalI3EnergyPDFSet
from skyllh.i3.backgroundpdf import (
    DataBackgroundI3SpatialPDF,
    DataBackgroundI3EnergyPDF
)
from skyllh.i3.pdfratio import (
    I3EnergySigSetOverBkgPDFRatioSpline
)
# Classes to define the spatial and energy PDF ratios.
from skyllh.core.pdfratio import (
    SpatialSigOverBkgPDFRatio,
    Skylab2SkylabPDFRatioFillMethod
)

from skyllh.i3.signal_generation import PointLikeSourceI3SignalGenerationMethod

# Analysis utilities.
from skyllh.core.analysis_utils import (
    pointlikesource_to_data_field_array
)

# Logging setup utilities.
from skyllh.core.debugging import (
    setup_logger,
    setup_console_handler,
    setup_file_handler
)

# Pre-defined public IceCube data samples.
from skyllh.datasets.i3 import data_samples

# Analysis specific classes for working with the public data.
from skyllh.analyses.i3.publicdata_ps.signal_generator import (
    PDSignalGenerator
)
from skyllh.analyses.i3.publicdata_ps.detsigyield import (
    PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod
)
from skyllh.analyses.i3.publicdata_ps.signalpdf import (
    PDSignalEnergyPDFSet
)
from skyllh.analyses.i3.publicdata_ps.backgroundpdf import (
    PDMCBackgroundI3EnergyPDF
)
from skyllh.analyses.i3.publicdata_ps.pdfratio import (
    PDPDFRatio
)


def psi_func(tdm, src_hypo_group_manager, fitparams):
    """Function to calculate the opening angle between the source position
    and the event's reconstructed position.
    """
    ra = tdm.get_data('ra')
    dec = tdm.get_data('dec')

    # Make the source position angles two-dimensional so the PDF value
    # can be calculated via numpy broadcasting automatically for several
    # sources. This is useful for stacking analyses.
    src_ra = tdm.get_data('src_array')['ra'][:, np.newaxis]
    src_dec = tdm.get_data('src_array')['dec'][:, np.newaxis]

    delta_dec = np.abs(dec - src_dec)
    delta_ra = np.abs(ra - src_ra)
    x = (
        (np.sin(delta_dec / 2.))**2. + np.cos(dec) *
        np.cos(src_dec) * (np.sin(delta_ra / 2.))**2.
    )

    # Handle possible floating precision errors.
    x[x < 0.] = 0.
    x[x > 1.] = 1.

    psi = (2.0*np.arcsin(np.sqrt(x)))

    # For now we support only a single source, hence return psi[0].
    return psi[0, :]


def TXS_location():
    src_ra = np.radians(77.358)
    src_dec = np.radians(5.693)
    return (src_ra, src_dec)


def create_analysis(
    rss,
    datasets,
    source,
    refplflux_Phi0=1,
    refplflux_E0=1e3,
    refplflux_gamma=2,
    ns_seed=10.0,
    gamma_seed=3,
    cache_dir='.',
    cap_ratio=False,
    compress_data=False,
    keep_data_fields=None,
    optimize_delta_angle=10,
    efficiency_mode=None,
    tl=None,
    ppbar=None
):
    """Creates the Analysis instance for this particular analysis.

    Parameters:
    -----------
    datasets : list of Dataset instances
        The list of Dataset instances, which should be used in the
        analysis.
    source : PointLikeSource instance
        The PointLikeSource instance defining the point source position.
    refplflux_Phi0 : float
        The flux normalization to use for the reference power law flux model.
    refplflux_E0 : float
        The reference energy to use for the reference power law flux model.
    refplflux_gamma : float
        The spectral index to use for the reference power law flux model.
    ns_seed : float
        Value to seed the minimizer with for the ns fit.
    gamma_seed : float | None
        Value to seed the minimizer with for the gamma fit. If set to None,
        the refplflux_gamma value will be set as gamma_seed.
    cache_dir : str
        The cache directory where to look for cached data, e.g. signal PDFs.
    compress_data : bool
        Flag if the data should get converted from float64 into float32.
    keep_data_fields : list of str | None
        List of additional data field names that should get kept when loading
        the data.
    optimize_delta_angle : float
        The delta angle in degrees for the event selection optimization methods.
    efficiency_mode : str | None
        The efficiency mode the data should get loaded with. Possible values
        are:

            - 'memory':
                The data will be load in a memory efficient way. This will
                require more time, because all data records of a file will
                be loaded sequentially.
            - 'time':
                The data will be loaded in a time efficient way. This will
                require more memory, because each data file gets loaded in
                memory at once.

        The default value is ``'time'``. If set to ``None``, the default
        value will be used.
    tl : TimeLord instance | None
        The TimeLord instance to use to time the creation of the analysis.
    ppbar : ProgressBar instance | None
        The instance of ProgressBar for the optional parent progress bar.

    Returns
    -------
    analysis : SpatialEnergyTimeIntegratedMultiDatasetSingleSourceAnalysis
        The Analysis instance for this analysis.
    """
    # Define the flux model.
    flux_model = PowerLawFlux(
        Phi0=refplflux_Phi0, E0=refplflux_E0, gamma=refplflux_gamma)

    # Define the fit parameter ns.
    fitparam_ns = FitParameter('ns', 0, 1e3, ns_seed)

    # Define the gamma fit parameter.
    fitparam_gamma = FitParameter(
        'gamma', valmin=1, valmax=5, initial=gamma_seed)

    # Define the detector signal efficiency implementation method for the
    # IceCube detector and this source and flux_model.
    # The sin(dec) binning will be taken by the implementation method
    # automatically from the Dataset instance.
    gamma_grid = fitparam_gamma.as_linear_grid(delta=0.1)
    detsigyield_implmethod = \
        PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
            gamma_grid)

    # Define the signal generation method.
    #sig_gen_method = PointLikeSourceI3SignalGenerationMethod()
    sig_gen_method = None

    # Create a source hypothesis group manager.
    src_hypo_group_manager = SourceHypoGroupManager(
        SourceHypoGroup(
            source, flux_model, detsigyield_implmethod, sig_gen_method))

    # Create a source fit parameter mapper and define the fit parameters.
    src_fitparam_mapper = SingleSourceFitParameterMapper()
    src_fitparam_mapper.def_fit_parameter(fitparam_gamma)

    # Define the test statistic.
    test_statistic = TestStatisticWilks()

    # Define the data scrambler with its data scrambling method, which is used
    # for background generation.
    data_scrambler = DataScrambler(UniformRAScramblingMethod())

    # Create background generation method.
    bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(data_scrambler)

    # Create the minimizer instance.
    minimizer = Minimizer(LBFGSMinimizerImpl())

    # Create the Analysis instance.
    analysis = Analysis(
        src_hypo_group_manager,
        src_fitparam_mapper,
        fitparam_ns,
        test_statistic,
        bkg_gen_method,
        sig_generator_cls=PDSignalGenerator
    )

    # Define the event selection method for pure optimization purposes.
    # We will use the same method for all datasets.
    event_selection_method = SpatialBoxEventSelectionMethod(
        src_hypo_group_manager, delta_angle=np.deg2rad(optimize_delta_angle))
    #event_selection_method = None

    # Add the data sets to the analysis.
    pbar = ProgressBar(len(datasets), parent=ppbar).start()
    for ds in datasets:
        # Load the data of the data set.
        data = ds.load_and_prepare_data(
            keep_fields=keep_data_fields,
            compress=compress_data,
            efficiency_mode=efficiency_mode,
            tl=tl)

        # Create a trial data manager and add the required data fields.
        tdm = TrialDataManager()
        tdm.add_source_data_field('src_array',
                                  pointlikesource_to_data_field_array)
        tdm.add_data_field('psi', psi_func)

        sin_dec_binning = ds.get_binning_definition('sin_dec')
        log_energy_binning = ds.get_binning_definition('log_energy')

        # Create the spatial PDF ratio instance for this dataset.
        spatial_sigpdf = RayleighPSFPointSourceSignalSpatialPDF(
            dec_range=np.arcsin(sin_dec_binning.range))
        spatial_bkgpdf = DataBackgroundI3SpatialPDF(
            data.exp, sin_dec_binning)
        spatial_pdfratio = SpatialSigOverBkgPDFRatio(
            spatial_sigpdf, spatial_bkgpdf)

        # Create the energy PDF ratio instance for this dataset.
        energy_sigpdfset = PDSignalEnergyPDFSet(
            ds=ds,
            src_dec=source.dec,
            flux_model=flux_model,
            fitparam_grid_set=gamma_grid,
            ppbar=ppbar
        )

        #smoothing_filter = BlockSmoothingFilter(nbins=1)
        #energy_bkgpdf = DataBackgroundI3EnergyPDF(
        #    data.exp, log_energy_binning, sin_dec_binning, smoothing_filter)

        bkg_pdf_pathfilename = ds.get_abs_pathfilename_list(
            ds.get_aux_data_definition('pdf_bkg_datafile'))[0]
        with open(bkg_pdf_pathfilename, 'rb') as f:
            bkg_pdf_data = pickle.load(f)
        energy_bkgpdf = PDMCBackgroundI3EnergyPDF(
            pdf_log10emu_sindecmu=bkg_pdf_data['pdf'],
            sindecmu_binning=bkg_pdf_data['sindecmu_binning'],
            log10emu_binning=bkg_pdf_data['log10emu_binning']
        )

        energy_pdfratio = PDPDFRatio(
            sig_pdf_set=energy_sigpdfset,
            bkg_pdf=energy_bkgpdf,
            cap_ratio=cap_ratio
        )

        pdfratios = [spatial_pdfratio, energy_pdfratio]

        analysis.add_dataset(
            ds, data, pdfratios, tdm, event_selection_method)

        pbar.increment()
    pbar.finish()

    analysis.llhratio = analysis.construct_llhratio(minimizer, ppbar=ppbar)

    # analysis.construct_signal_generator()

    return analysis


if(__name__ == '__main__'):
    p = argparse.ArgumentParser(
        description='Calculates TS for a given source location using the '
        '10-year public point source sample.',
        formatter_class=argparse.RawTextHelpFormatter
    )
    p.add_argument(
        '--dec',
        default=23.8,
        type=float,
        help='The source declination in degrees.'
    )
    p.add_argument(
        '--ra',
        default=216.76,
        type=float,
        help='The source right-ascention in degrees.'
    )
    p.add_argument(
        '--gamma-seed',
        default=3,
        type=float,
        help='The seed value of the gamma fit parameter.'
    )
    p.add_argument(
        '--data_base_path',
        default=None,
        type=str,
        help='The base path to the data samples (default=None)'
    )
    p.add_argument(
        '--pdf-seed',
        default=1,
        type=int,
        help='The random number generator seed for generating the '
             'signal PDF.'
    )
    p.add_argument(
        '--seed',
        default=1,
        type=int,
        help='The random number generator seed for the likelihood '
             'minimization.'
    )
    p.add_argument(
        '--ncpu',
        default=1,
        type=int,
        help='The number of CPUs to utilize where parallelization is possible.'
    )
    p.add_argument(
        '--cache-dir',
        default='.',
        type=str,
        help='The cache directory to look for cached data, e.g. signal PDFs.')
    p.add_argument(
        '--cap-ratio',
        action='store_true',
        help='Switch to cap the energy PDF ratio.')
    p.set_defaults(cap_ratio=False)
    args = p.parse_args()

    # Setup `skyllh` package logging.
    # To optimize logging set the logging level to the lowest handling level.
    setup_logger('skyllh', logging.DEBUG)
    log_format = '%(asctime)s %(processName)s %(name)s %(levelname)s: '\
                 '%(message)s'
    setup_console_handler('skyllh', logging.INFO, log_format)
    setup_file_handler('skyllh', 'debug.log',
        log_level=logging.DEBUG,
        log_format=log_format)

    CFG['multiproc']['ncpu'] = args.ncpu

    sample_seasons = [
        #('PublicData_10y_ps', 'IC40'),
        #('PublicData_10y_ps', 'IC59'),
        #('PublicData_10y_ps', 'IC79'),
        #('PublicData_10y_ps', 'IC86_I'),
        ('PublicData_10y_ps', 'IC86_II'),
        #('PublicData_10y_ps', 'IC86_II-VII')
    ]

    datasets = []
    for (sample, season) in sample_seasons:
        # Get the dataset from the correct dataset collection.
        dsc = data_samples[sample].create_dataset_collection(
            args.data_base_path)
        datasets.append(dsc.get_dataset(season))

    # Define a random state service.
    rss_pdf = RandomStateService(args.pdf_seed)
    rss = RandomStateService(args.seed)
    # Define the point source.
    source = PointLikeSource(np.deg2rad(args.ra), np.deg2rad(args.dec))
    print('source: ', str(source))

    tl = TimeLord()

    with tl.task_timer('Creating analysis.'):
        ana = create_analysis(
            rss_pdf,
            datasets,
            source,
            cache_dir=args.cache_dir,
            cap_ratio=args.cap_ratio,
            gamma_seed=args.gamma_seed,
            tl=tl)

    with tl.task_timer('Unblinding data.'):
        (TS, fitparam_dict, status) = ana.unblind(rss)

    print('TS = %g' % (TS))
    print('ns_fit = %g' % (fitparam_dict['ns']))
    print('gamma_fit = %g' % (fitparam_dict['gamma']))


    # Generate some signal events.
    #ana.construct_signal_generator()
    #with tl.task_timer('Generating signal events.'):
    #    (n_sig, signal_events_dict) =\
    #        ana.sig_generator.generate_signal_events(rss, 100)

    #trials = ana.do_trials(
    #       rss, 100, mean_n_sig=20
    #)

    #print('n_sig: %d'%n_sig)
    #print('signal datasets: '+str(signal_events_dict.keys()))


    print(tl)



# -*- coding: utf-8 -*-

"""The llhratio module provides classes implementing the log-likelihood ratio
functions. In general these should be detector independent, because they
implement the pure math of the log-likelihood ratio function.
"""

from __future__ import division

import abc
import numpy as np

from skyllh.core.config import CFG
from skyllh.core.debugging import get_logger
from skyllh.core.py import (
    classname,
    int_cast,
    issequence,
    issequenceof,
    float_cast
)
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.core.trialdata import TrialDataManager
from skyllh.core.detsigyield import DetSigYield
from skyllh.core.minimizer import (
    Minimizer,
    NR1dNsMinimizerImpl,
    NRNsScan2dMinimizerImpl
)
from skyllh.core.parameters import (
    SourceFitParameterMapper,
    SingleSourceFitParameterMapper,
    MultiSourceFitParameterMapper
)
from skyllh.core.pdfratio import (
    PDFRatio,
    SingleSourcePDFRatioArrayArithmetic
)
from skyllh.core.timing import TaskTimer
from skyllh.physics.source import SourceModel


logger = get_logger(__name__)


class LLHRatio(object, metaclass=abc.ABCMeta):
    """Abstract base class for a log-likelihood (LLH) ratio function.
    """

    def __init__(self, minimizer):
        """Creates a new LLH ratio function instance.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        """
        super(LLHRatio, self).__init__()

        self.minimizer = minimizer

    @property
    def minimizer(self):
        """The Minimizer instance used to minimize the negative of the
        log-likelihood ratio function.
        """
        return self._minimizer
    @minimizer.setter
    def minimizer(self, minimizer):
        if(not isinstance(minimizer, Minimizer)):
            raise TypeError('The minimizer property must be an instance '
                'of Minimizer!')
        self._minimizer = minimizer

    def initialize_for_new_trial(self, tl=None):
        """This method will be called by the Analysis class after new trial data
        has been initialized to the trial data manager. Derived classes can make
        use of this call hook to perform LLHRatio specific trial initialization.

        Parameters
        ----------
        tl : TimeLord | None
            The optional TimeLord instance to use for timing measurements.
        """
        pass

    @abc.abstractmethod
    def evaluate(self, fitparam_values, tl=None):
        """This method evaluates the LLH ratio function for the given set of
        fit parameter values.

        Parameters
        ----------
        fitparam_values : numpy 1D ndarray
            The ndarray holding the current values of the (global) fit
            parameters.
        tl : TimeLord | None
            The optional TimeLord instance to use for measuring timeing.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value.
        grads : (N_fitparams,)-shaped 1D ndarray
            The ndarray holding the gradient value for each (global) fit
            parameter.
        """
        pass

    def maximize(self, rss, fitparamset, tl=None):
        """Maximize the log-likelihood ratio function, by using the ``evaluate``
        method.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance to draw random numbers from.
            This is needed to generate random parameter initial values.
        fitparamset : FitParameterSet instance
            The instance of FitParameterSet holding the global fit parameter
            definitions used in the maximization process.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to time the
            maximization process.

        Returns
        -------
        log_lambda_max : float
            The (maximum) value of the log-likelihood ratio (log_lambda)
            function for the best fit parameter values.
        fitparam_values : (N_fitparam)-shaped 1D ndarray
            The ndarray holding the global fit parameter values.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        tracing = CFG['debugging']['enable_tracing']

        # Define the negative llhratio function, that will get minimized.
        self_evaluate = self.evaluate
        def negative_llhratio_func(fitparam_values, func_stats, tl=None):
            func_stats['n_calls'] += 1
            with TaskTimer(tl, 'Evaluate llh-ratio function.'):
                (f, grads) = self_evaluate(fitparam_values, tl=tl)
                if(tracing): logger.debug(
                    'LLH-ratio func value f={:g}, grads={}'.format(
                        f, str(grads)))
            return (-f, -grads)

        minimize_kwargs = {'func_provides_grads': True}

        func_stats = {'n_calls': 0}
        with TaskTimer(tl, 'Minimize -llhratio function.'):
            (fitparam_values, fmin, status) = self._minimizer.minimize(
                rss, fitparamset, negative_llhratio_func, args=(func_stats,tl),
                kwargs=minimize_kwargs)
        log_lambda_max = -fmin
        status['n_llhratio_func_calls'] = func_stats['n_calls']

        logger.debug(
            'Maximized LLH ratio function with {:d} calls'.format(
                status['n_llhratio_func_calls']))

        return (log_lambda_max, fitparam_values, status)


class TCLLHRatio(LLHRatio, metaclass=abc.ABCMeta):
    """Abstract base class for a log-likelihood (LLH) ratio function with two
    components, i.e. signal and background.
    """

    def __init__(self, minimizer, mean_n_sig_0):
        """Creates a new two-component LLH ratio function instance.

        Parameters
        ----------
        mean_n_sig_0 : float
            The fixed mean number of signal events for the null-hypothesis.
        """
        super(TCLLHRatio, self).__init__(minimizer)

        self.mean_n_sig_0 = mean_n_sig_0

    @property
    def mean_n_sig_0(self):
        """The parameter value for the mean number of signal events of the
        null-hypothesis.
        """
        return self._mean_n_sig_0
    @mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        v = float_cast(v, 'The mean_n_sig_0 property must be castable to a '
            'float value!')
        self._mean_n_sig_0 = v

    @abc.abstractmethod
    def calculate_ns_grad2(self, fitparam_values):
        """This method is supposed to calculate the second derivative of the
        log-likelihood ratio function w.r.t. the fit parameter ns, the number
        of signal events in the data set.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams,)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        pass

    def maximize(self, rss, fitparamset, tl=None):
        """Maximizes this log-likelihood ratio function, by minimizing its
        negative.
        This method has a special implementation when a 1D Newton-Rapson
        minimizer is used. In that case only the first and second derivative
        of the log-likelihood ratio function is calculated.

        Parameters
        ----------
        rss : RandomStateService instance
            The RandomStateService instance that should be used to draw random
            numbers from. It is used by the minimizer to generate random
            fit parameter initial values.
        fitparamset : FitParameterSet instance
            The instance of FitParameterSet holding the global fit parameter
            definitions used in the maximization process.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to time the
            maximization of the LLH-ratio function.

        Returns
        -------
        log_lambda_max : float
            The (maximum) value of the log-likelihood ratio (log_lambda)
            function for the best fit parameter values.
        fitparam_values : (N_fitparam)-shaped 1D ndarray
            The ndarray holding the global fit parameter values.
        status : dict
            The dictionary with status information about the maximization
            process, i.e. from the minimizer.
        """
        if(isinstance(self._minimizer.minimizer_impl, NR1dNsMinimizerImpl) or
           isinstance(self._minimizer.minimizer_impl, NRNsScan2dMinimizerImpl)):
            # Define the negative llhratio function, that will get minimized
            # when using the Newton-Rapson 1D minimizer for llhratio functions
            # depending solely on ns.
            self__evaluate = self.evaluate
            self__calculate_ns_grad2 = self.calculate_ns_grad2
            def negative_llhratio_func_nr1d_ns(fitparam_values, tl):
                with TaskTimer(tl, 'Evaluate llh-ratio function.'):
                    (f, grads) = self__evaluate(fitparam_values, tl=tl)
                with TaskTimer(tl, 'Calculate 2nd derivative of llh-ratio '
                        'function w.r.t. ns'):
                    grad2_ns = self__calculate_ns_grad2(fitparam_values)

                return (-f, -grads[0], -grad2_ns)

            (fitparam_values, fmin, status) = self._minimizer.minimize(
                rss, fitparamset, negative_llhratio_func_nr1d_ns, args=(tl,))
            log_lambda_max = -fmin

            return (log_lambda_max, fitparam_values, status)

        return super(TCLLHRatio, self).maximize(rss, fitparamset, tl=tl)


class SingleDatasetTCLLHRatio(TCLLHRatio, metaclass=abc.ABCMeta):
    """Abstract base class for a log-likelihood (LLH) ratio function with two
    components, i.e. signal and background, for a single data set.
    """

    def __init__(
            self, minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            mean_n_sig_0):
        """Creates a new two-component LLH ratio function instance for a single
        data set.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance that defines the source
            hypotheses.
        src_fitparam_mapper : SourceFitParameterMapper
            The instance of SourceFitParameterMapper defining the global fit
            parameters and their mapping to the source fit parameters.
            The order of the fit parameters defines the order of the fit values
            during the maximization process of the log-likelihood-ratio
            function. The names of the source fit parameters must coincide with
            the signal fit parameter names of the PDF instances.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        mean_n_sig_0 : float
            The fixed mean number of signal events for the null-hypothesis.
        """
        super(SingleDatasetTCLLHRatio, self).__init__(
            minimizer, mean_n_sig_0)

        self.src_hypo_group_manager = src_hypo_group_manager
        self.src_fitparam_mapper = src_fitparam_mapper
        self.tdm = tdm

        # Calculate the data fields that solely depend on source parameters.
        self._tdm.calculate_source_data_fields(src_hypo_group_manager)

    @property
    def src_hypo_group_manager(self):
        """The SourceHypoGroupManager instance that defines the source
        hypotheses.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError('The src_hypo_group_manager property must be an '
                'instance of SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def src_fitparam_mapper(self):
        """The SourceFitParameterMapper instance defining the global fit
        parameters and their mapping to the source fit parameters.
        """
        return self._src_fitparam_mapper
    @src_fitparam_mapper.setter
    def src_fitparam_mapper(self, mapper):
        if(not isinstance(mapper, SourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper property must be an '
                'instance of SourceFitParameterMapper!')
        self._src_fitparam_mapper = mapper

    @property
    def tdm(self):
        """The TrialDataManager instance that holds the trial event data and
        additional data fields for this LLH ratio function.
        """
        return self._tdm
    @tdm.setter
    def tdm(self, manager):
        if(not isinstance(manager, TrialDataManager)):
            raise TypeError('The tdm property must be an instance of '
                'TrialDataManager!')
        self._tdm = manager

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the source hypothesis group manager of this two-component LLH
        ratio function.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self._tdm.change_source_hypo_group_manager(src_hypo_group_manager)


class ZeroSigH0SingleDatasetTCLLHRatio(SingleDatasetTCLLHRatio):
    """This class implements a two-component (TC) log-likelihood (LLH) ratio
    function for a single data assuming zero signal for the null-hypothesis.
    The log-likelihood-ratio function uses a list of independent PDF ratio
    instances.
    """
    # The (1 + alpha)-threshold float value for which the log-likelihood ratio
    # function of a single event should get approximated by a Taylor expansion.
    # This is to prevent a divergence of the log-function for each event, where
    # (1 + alpha_i) < (1 + alpha).
    # This setting is implemented as a class type member instead of a class
    # instance member, because it is supposed to be the same for all instances.
    _one_plus_alpha = 1e-3

    def __init__(
            self, minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            pdfratios):
        """Constructor of the two-component log-likelihood ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance that defines the source
            hypotheses.
        src_fitparam_mapper : SourceFitParameterMapper
            The instance of SourceFitParameterMapper defining the global fit
            parameters and their mapping to the source fit parameters.
            The order of the fit parameters defines the order of the fit values
            during the maximization process of the log-likelihood-ratio
            function. The names of the source fit parameters must coincide with
            the signal fit parameter names of the PDF ratio instances.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        pdfratios : sequence of PDFRatio
            The sequence of PDFRatio instances. A PDFRatio instance might depend
            on none, one, or several fit parameters.
        """
        super(ZeroSigH0SingleDatasetTCLLHRatio, self).__init__(
            minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            mean_n_sig_0=0)

        self.pdfratio_list = pdfratios

        # Define cache variables for evaluate method to store values needed for
        # a possible calculation of the second derivative w.r.t. ns of the
        # log-likelihood ratio function.
        self._cache_fitparam_values = None
        self._cache_nsgrad_i = None

    @SingleDatasetTCLLHRatio.mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        SingleDatasetTCLLHRatio.mean_n_sig_0.fset(self, v)
        if(self._mean_n_sig_0 != 0):
            raise ValueError('The %s class is only valid for '
                'mean_n_sig_0 = 0!'%(classname(self)))

    @property
    def pdfratio_list(self):
        """The list of PDFRatio instances.
        """
        return self._pdfratio_list
    @pdfratio_list.setter
    def pdfratio_list(self, seq):
        if(not issequenceof(seq, PDFRatio)):
            raise TypeError('The pdfratio_list property must be a sequence of '
                'PDFRatio instances!')
        self._pdfratio_list = list(seq)

    def calculate_log_lambda_and_grads(self, fitparam_values, N, ns, Xi, dXi_ps):
        """Calculates the log(Lambda) value and its gradient for each global fit
        parameter. This calculation is source and detector independent.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.
            These numbers are used as cache key to validate the ``nsgrad_i``
            values for the given fit parameter values for a possible later
            calculation of the second derivative w.r.t. ns of the log-likelihood
            ratio function.
        N : int
            The total number of events.
        ns : float
            The current fit parameter value for ns.
        Xi : numpy (n_selected_events,)-shaped 1D ndarray
            The X value of each selected event.
        dXi_ps : numpy (N_fitparams,n_selected_events)-shaped 2D ndarray
            The derivative value for each fit parameter ps of each event's X
            value.

        Returns
        -------
        log_lambda : float
            The value of the log-likelihood ratio function.
        grads : 1D numpy (N_fitparams+1,)-shaped ndarray
            The gradient value of log_lambda for each fit parameter.
            The first element is the gradient for ns.
        """
        tracing = CFG['debugging']['enable_tracing']

        # Get the number of selected events.
        Nprime = len(Xi)

        if(tracing):
            logger.debug(
                'N={:d}, Nprime={:d}'.format(
                    N, Nprime))

        one_plus_alpha = ZeroSigH0SingleDatasetTCLLHRatio._one_plus_alpha

        alpha = one_plus_alpha - 1
        alpha_i = ns*Xi

        # Create a mask for events which have a stable non-diverging
        # log-function argument, and an inverted mask thereof.
        stablemask = alpha_i > alpha
        unstablemask = ~stablemask
        if(tracing):
            logger.debug(
                '# of events doing Taylor expansion for (unstable events): '
                '{:d}'.format(
                    np.count_nonzero(unstablemask)))

        # Allocate memory for the log_lambda_i values.
        log_lambda_i = np.empty_like(alpha_i, dtype=np.float64)

        # Calculate the log_lambda_i value for the numerical stable events.
        log_lambda_i[stablemask] = np.log1p(alpha_i[stablemask])
        # Calculate the log_lambda_i value for the numerical unstable events.
        tildealpha_i = (alpha_i[unstablemask] - alpha) / one_plus_alpha
        log_lambda_i[unstablemask] = np.log1p(alpha) + tildealpha_i - 0.5*tildealpha_i**2

        # Calculate the log_lambda value and account for pure background events.
        log_lambda = np.sum(log_lambda_i) + (N - Nprime)*np.log1p(-ns/N)

        # Calculate the gradient for each fit parameter.
        grads = np.empty((dXi_ps.shape[0]+1,), dtype=np.float64)

        # Pre-calculate value that is used twice for the gradients of the
        # numerical stable events.
        one_over_one_plus_alpha_i_stablemask = 1 / (1 + alpha_i[stablemask])

        # For ns.
        nsgrad_i = np.empty_like(alpha_i, dtype=np.float64)
        nsgrad_i[stablemask] = Xi[stablemask] * one_over_one_plus_alpha_i_stablemask
        nsgrad_i[unstablemask] = (1 - tildealpha_i)*Xi[unstablemask] / one_plus_alpha
        # Cache the nsgrad_i values for a possible later calculation of the
        # second derivative w.r.t. ns of the log-likelihood ratio function.
        # Note: We create a copy of the fitparam_values array here to make sure
        #       that the values don't get changed outside this method before the
        #       calculate_ns_grad2 method is called.
        self._cache_fitparam_values = fitparam_values.copy()
        self._cache_nsgrad_i = nsgrad_i
        # Calculate the first derivative w.r.t. ns.
        grads[0] = np.sum(nsgrad_i) - (N - Nprime)/(N - ns)

        # For each other fit parameter.
        # For all numerical stable events.
        grads[1:] = np.sum(ns * one_over_one_plus_alpha_i_stablemask * dXi_ps[:,stablemask], axis=1)
        # For all numerical unstable events.
        grads[1:] += np.sum(ns*(1 - tildealpha_i)*dXi_ps[:,unstablemask] / one_plus_alpha, axis=1)

        return (log_lambda, grads)

    def calculate_ns_grad2(self, fitparam_values):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.
        This method tries to use cached values for the first derivative
        w.r.t. ns of the log-likelihood ratio function for the individual
        events. If cached values don't exist or do not match the given fit
        parameter values, they will get calculated automatically by calling the
        evaluate method with the given fit parameter values.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        # Check if the cached nsgrad_i values match the given fitparam_values.
        if((self._cache_fitparam_values is None) or
           (not np.all(self._cache_fitparam_values == fitparam_values))):
            # Calculate the cache values by evaluating the log-likelihood ratio
            # function.
            self.evaluate(fitparam_values)

        ns = fitparam_values[0]
        Nprime = self._tdm.n_selected_events
        N = Nprime + self._tdm.n_pure_bkg_events

        nsgrad2 = -np.sum(self._cache_nsgrad_i**2) - (N - Nprime)/(N - ns)**2

        return nsgrad2


class SingleSourceZeroSigH0SingleDatasetTCLLHRatio(
        ZeroSigH0SingleDatasetTCLLHRatio):
    """This class implements a 2-component, i.e. signal and background,
    log-likelihood ratio function for a single data set. The
    log-likelihood-ratio function assumes a zero signal null-hypothesis and uses
    a list of independent PDFRatio instances assuming a single source.
    """
    def __init__(
            self, minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            pdfratios):
        """Constructor for creating a 2-component, i.e. signal and background,
        log-likelihood ratio function assuming a single source.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance that defines the source
            hypotheses.
        src_fitparam_mapper : SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
            The order of the fit parameters defines the order of the fit values
            during the maximization process.
            The names of the source fit parameters must coincide with the signal
            fit parameter names of the PDF ratio objects.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        pdfratios : list of PDFRatio
            The list of PDFRatio instances. A PDFRatio instance might depend on
            none, one, or several fit parameters.
        """
        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        super(SingleSourceZeroSigH0SingleDatasetTCLLHRatio, self).__init__(
            minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            pdfratios)

        # Construct a PDFRatio array arithmetic object specialized for a single
        # source. This will pre-calculate the PDF ratio values for all PDF ratio
        # instances, which do not depend on any fit parameters.
        self._pdfratioarray = SingleSourcePDFRatioArrayArithmetic(
            self._pdfratio_list,
            self._src_fitparam_mapper.fitparamset.fitparam_list)

    def initialize_for_new_trial(self, tl=None):
        """Initializes the log-likelihood ratio function for a new trial.

        Parameters
        ----------
        tl : TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        self._pdfratioarray.initialize_for_new_trial(self._tdm)

    def evaluate(self, fitparam_values, tl=None):
        """Evaluates the log-likelihood ratio function for the given set of
        data events.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.
        tl : TimeLord instance | None
            The optional TimeLord instance to measure the timing of evaluating
            the LLH ratio function.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value.
        grads : (N_fitparams+1,)-shaped 1D ndarray
            The ndarray holding the gradient value of log_lambda for each fit
            parameter and ns.
            The first element is the gradient for ns.
        """
        tracing = CFG['debugging']['enable_tracing']

        # Define local variables to avoid (.)-lookup procedure.
        tdm = self._tdm
        pdfratioarray = self._pdfratioarray

        ns = fitparam_values[0]

        N = tdm.n_events

        # Create the fitparams dictionary with the fit parameter names and
        # values.
        with TaskTimer(tl, 'Create fitparams dictionary.'):
            fitparams = self._src_fitparam_mapper.get_src_fitparams(
                fitparam_values[1:])

        # Calculate the data fields that depend on fit parameter values.
        with TaskTimer(tl, 'Calc fit param dep data fields.'):
            tdm.calculate_fitparam_data_fields(
                self._src_hypo_group_manager, fitparams)

        # Calculate the PDF ratio values of all PDF ratio objects, which depend
        # on any fit parameter.
        with TaskTimer(tl, 'Calc pdfratio values.'):
            pdfratioarray.calculate_pdfratio_values(tdm, fitparams, tl=tl)

        # Calculate the product of all the PDF ratio values for each (selected)
        # event.
        with TaskTimer(tl, 'Calc pdfratio value product Ri'):
            Ri = pdfratioarray.get_ratio_product()

        # Calculate Xi for each (selected) event.
        Xi = (Ri - 1.) / N
        if(tracing):
            logger.debug('dtype(Xi)={:s}'.format(str(Xi.dtype)))

        # Calculate the gradients of Xi for each fit parameter (without ns).
        dXi_ps = np.empty((len(fitparam_values)-1,len(Xi)), dtype=np.float64)
        for (idx, fitparam_value) in enumerate(fitparam_values[1:]):
            fitparam_name = self._src_fitparam_mapper.get_src_fitparam_name(idx)

            dRi = np.zeros((len(Xi),), dtype=np.float64)
            for (num_k) in np.arange(len(pdfratioarray._pdfratio_list)):
                # Get the PDFRatio instance from which we need the derivative from.
                pdfratio = pdfratioarray.get_pdfratio(num_k)
                # Calculate the derivative of Ri.
                dRi += pdfratio.get_gradient(tdm, fitparams, fitparam_name) * pdfratioarray.get_ratio_product(excluded_idx=num_k)

            # Calculate the derivative of Xi w.r.t. the fit parameter.
            dXi_ps[idx] = dRi / N

        if(tracing):
            logger.debug(
                '{:s}.evaluate: N={:d}, Nprime={:d}, ns={:.3f}, '.format(
                    classname(self), N, len(Xi), ns))

        with TaskTimer(tl, 'Calc logLamds and grads'):
            (log_lambda, grads) = self.calculate_log_lambda_and_grads(
                fitparam_values, N, ns, Xi, dXi_ps)

        return (log_lambda, grads)


class MultiSourceZeroSigH0SingleDatasetTCLLHRatio(
        SingleSourceZeroSigH0SingleDatasetTCLLHRatio):
    """This class implements a 2-component, i.e. signal and background,
    log-likelihood ratio function for a single data set assuming zero signal for
    the null-hypothesis. It uses a list of independent PDFRatio instances
    assuming multiple sources (stacking).
    """
    def __init__(
            self, minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            pdfratios, detsigyields):
        """Constructor for creating a 2-component, i.e. signal and background,
        log-likelihood ratio function assuming a single source.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance that defines the source
            hypotheses.
        src_fitparam_mapper : SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
            The order of the fit parameters defines the order of the fit values
            during the maximization process.
            The names of the source fit parameters must coincide with the signal
            fit parameter names of the PDF ratio objects.
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data and
            additional data fields for this LLH ratio function.
        pdfratios : list of PDFRatio
            The list of PDFRatio instances. A PDFRatio instance might depend on
            none, one, or several fit parameters.
        detsigyields : (N_source_hypo_groups,)-shaped 1D ndarray of DetSigYield
                instances
            The collection of DetSigYield instances for each source hypothesis
            group.
        """
        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        super(MultiSourceZeroSigH0SingleDatasetTCLLHRatio, self).__init__(
            minimizer, src_hypo_group_manager, src_fitparam_mapper, tdm,
            pdfratios)

        # Construct a PDFRatio array arithmetic object specialized for a single
        # source. This will pre-calculate the PDF ratio values for all PDF ratio
        # instances, which do not depend on any fit parameters.
        self._pdfratioarray = SingleSourcePDFRatioArrayArithmetic(
            self._pdfratio_list,
            self._src_fitparam_mapper.fitparamset.fitparam_list)

        self._calc_source_weights = MultiPointSourcesRelSourceWeights(
            src_hypo_group_manager, src_fitparam_mapper, detsigyields)

    def evaluate(self, fitparam_values, tl=None):
        """Evaluates the log-likelihood ratio function for the given set of
        data events.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.
        tl : TimeLord instance | None
            The optional TimeLord instance to measure the timing of evaluating
            the LLH ratio function.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value.
        grads : (N_fitparams+1,)-shaped 1D ndarray
            The ndarray holding the gradient value of log_lambda for each fit
            parameter and ns.
            The first element is the gradient for ns.
        """
        _src_w, _src_w_grads = self._calc_source_weights(
                fitparam_values)
        self._tdm.get_data('src_array')['src_w'] = _src_w
        if _src_w_grads is not None:
            self._tdm.get_data('src_array')['src_w_grad'] = _src_w_grads.flatten()
        else:
            self._tdm.get_data('src_array')['src_w_grad'] = np.zeros_like(_src_w)

        (log_lambda, grads) = super(
            MultiSourceZeroSigH0SingleDatasetTCLLHRatio, self).evaluate(
                fitparam_values, tl)

        return (log_lambda, grads)


class DatasetSignalWeights(object, metaclass=abc.ABCMeta):
    """Abstract base class for a dataset signal weight calculator class.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, detsigyields):
        """Base class constructor.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of the SourceHypoGroupManager managing the source
            hypothesis groups.
        src_fitparam_mapper : SourceFitParameterMapper
            The SourceFitParameterMapper instance that defines the global fit
            parameters and their mapping to the source fit parameters.
        detsigyields : 2D (N_source_hypo_groups,N_datasets)-shaped ndarray of
                     DetSigYield instances
            The collection of DetSigYield instances for each
            dataset and source group combination. The detector signal yield
            instances are used to calculate the dataset signal weight factors.
            The order must follow the definition order of the log-likelihood
            ratio functions, i.e. datasets, and the definition order of the
            source hypothesis groups.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self.src_fitparam_mapper = src_fitparam_mapper
        self.detsigyield_arr = detsigyields

        if(self._detsigyield_arr.shape[0] != self._src_hypo_group_manager.n_src_hypo_groups):
            raise ValueError('The detsigyields array must have the same number '
                'of source hypothesis groups as the source hypothesis group '
                'manager defines!')

        # Pre-convert the source list of each source hypothesis group into a
        # source array needed for the detector signal yield evaluation.
        # Since all the detector signal yield instances must be of the same
        # kind for each dataset, we can just use the one of the first dataset of
        # each source hypothesis group.
        self._src_arr_list = self._create_src_arr_list(
            self._src_hypo_group_manager, self._detsigyield_arr)

    def _create_src_arr_list(self, src_hypo_group_manager, detsigyield_arr):
        """Pre-convert the source list of each source hypothesis group into a
        source array needed for the detector signal yield evaluation.
        Since all the detector signal yield instances must be of the same
        kind for each dataset, we can just use the one of the first dataset of
        each source hypothesis group.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the sources.

        detsigyield_arr : 2D (N_source_hypo_groups,N_datasets)-shaped ndarray of
                        DetSigYield instances
            The collection of DetSigYield instances for each dataset and source
            group combination.
        Returns
        -------
        src_arr_list : list of numpy record ndarrays
            The list of the source numpy record ndarrays, one for each source
            hypothesis group, which is needed by the detector signal yield
            instance.
        """
        src_arr_list = []
        for (gidx, src_hypo_group) in enumerate(src_hypo_group_manager.src_hypo_group_list):
            src_arr_list.append(
                detsigyield_arr[gidx,0].source_to_array(src_hypo_group.source_list)
            )

        return src_arr_list

    @property
    def src_hypo_group_manager(self):
        """The instance of SourceHypoGroupManager, which defines the source
        hypothesis groups.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError('The src_hypo_group_manager property must be an '
                'instance of SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def src_fitparam_mapper(self):
        """The SourceFitParameterMapper instance defining the global fit
        parameters and their mapping to the source fit parameters.
        """
        return self._src_fitparam_mapper
    @src_fitparam_mapper.setter
    def src_fitparam_mapper(self, mapper):
        if(not isinstance(mapper, SourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper property must be an '
                'instance of SourceFitParameterMapper!')
        self._src_fitparam_mapper = mapper

    @property
    def detsigyield_arr(self):
        """The 2D (N_source_hypo_groups,N_datasets)-shaped ndarray of
        DetSigYield instances.
        """
        return self._detsigyield_arr
    @detsigyield_arr.setter
    def detsigyield_arr(self, detsigyields):
        if(not isinstance(detsigyields, np.ndarray)):
            raise TypeError('The detsigyield_arr property must be an instance '
                'of numpy.ndarray!')
        if(detsigyields.ndim != 2):
            raise ValueError('The detsigyield_arr property must be a '
                'numpy.ndarray with 2 dimensions!')
        if(not issequenceof(detsigyields.flat, DetSigYield)):
            raise TypeError('The detsigyield_arr property must contain '
                'DetSigYield instances, one for each source hypothesis group '
                'and dataset combination!')
        self._detsigyield_arr = detsigyields

    @property
    def n_datasets(self):
        """(read-only) The number of datasets this DatasetSignalWeights instance
        is for.
        """
        return self._detsigyield_arr.shape[1]

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the SourceHypoGroupManager instance of this
        DatasetSignalWeights instance. This will also recreate the internal
        source numpy record arrays needed for the detector signal efficiency
        calculation.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance, that should be used for
            this dataset signal weights instance.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self._src_arr_list = self._create_src_arr_list(
            self._src_hypo_group_manager, self._detsigyield_arr)

    @abc.abstractmethod
    def __call__(self, fitparam_values):
        """This method is supposed to calculate the dataset signal weights and
        their gradients.

        Parameters
        ----------
        fitparam_values : (N_fitparams+1,)-shaped 1D numpy ndarray
            The ndarray holding the current values of the fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        f : (N_datasets,)-shaped 1D ndarray
            The dataset signal weight factor for each dataset.
        f_grads : (N_datasets,N_fitparams)-shaped 2D ndarray
            The gradients of the dataset signal weight factors, one for each
            fit parameter.
        """
        pass


class SingleSourceDatasetSignalWeights(DatasetSignalWeights):
    """This class calculates the dataset signal weight factors for each dataset
    assuming a single source.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, detsigyields):
        """Constructs a new DatasetSignalWeights instance assuming a single
        source.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of the SourceHypoGroupManager managing the source
            hypothesis groups.
        src_fitparam_mapper : SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
        detsigyields : 2D (N_source_hypo_groups,N_datasets)-shaped ndarray of
                     DetSigYield instances
            The collection of DetSigYield instances for each
            dataset and source group combination. The detector signal yield
            instances are used to calculate the dataset signal weight factors.
            The order must follow the definition order of the log-likelihood
            ratio functions, i.e. datasets, and the definition order of the
            source hypothesis groups.
        """

        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        # Convert sequence into a 2D numpy array.
        detsigyields = np.atleast_2d(detsigyields)

        super(SingleSourceDatasetSignalWeights, self).__init__(
            src_hypo_group_manager, src_fitparam_mapper, detsigyields)

    def __call__(self, fitparam_values):
        """Calculates the dataset signal weight and its fit parameter gradients
        for each dataset.

        Parameters
        ----------
        fitparam_values : (N_fitparams+1,)-shaped 1D numpy ndarray
            The ndarray holding the current values of the fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        f : (N_datasets,)-shaped 1D ndarray
            The dataset signal weight factor for each dataset.
        f_grads : (N_datasets,N_fitparams)-shaped 2D ndarray | None
            The gradients of the dataset signal weight factors, one for each
            fit parameter. None is returned if there are no fit parameters
            beside ns.
        """
        fitparams_arr = self._src_fitparam_mapper.get_fitparams_array(fitparam_values[1:])

        N_datasets = self.n_datasets
        N_fitparams = self._src_fitparam_mapper.n_global_fitparams

        Y = np.empty((N_datasets,), dtype=np.float64)
        if(N_fitparams > 0):
            Y_grads = np.empty((N_datasets, N_fitparams), dtype=np.float64)

        # Loop over the detector signal efficiency instances for the first and
        # only source hypothesis group.
        for (j, detsigyield) in enumerate(self._detsigyield_arr[0]):
            (Yj, Yj_grads) = detsigyield(self._src_arr_list[0], fitparams_arr)
            # Store the detector signal yield and its fit parameter
            # gradients for the first and only source (element 0).
            Y[j] = Yj[0]
            if(N_fitparams > 0):
                if Yj_grads is None:
                    Y_grads[j] = np.zeros_like(Yj[0])
                else:
                    Y_grads[j] = Yj_grads[0]

        # sumj_Y is a scalar.
        sumj_Y = np.sum(Y, axis=0)

        # f is a (N_datasets,)-shaped 1D ndarray.
        f = Y/sumj_Y

        # f_grads is a (N_datasets, N_fitparams)-shaped 2D ndarray.
        if(N_fitparams > 0):
            # sumj_Y_grads is a (N_fitparams,)-shaped 1D array.
            sumj_Y_grads = np.sum(Y_grads, axis=0)
            f_grads = (Y_grads*sumj_Y - Y[...,np.newaxis]*sumj_Y_grads) / sumj_Y**2
        else:
            f_grads = None

        return (f, f_grads)


class MultiSourceDatasetSignalWeights(SingleSourceDatasetSignalWeights):
    """This class calculates the dataset signal weight factors for each dataset
    assuming multiple sources.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, detsigyields):
        """Constructs a new DatasetSignalWeights instance assuming multiple
        sources.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of the SourceHypoGroupManager managing the source
            hypothesis groups.
        src_fitparam_mapper : SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
        detsigyields : 2D (N_source_hypo_groups,N_datasets)-shaped ndarray of
                     DetSigYield instances
            The collection of DetSigYield instances for each
            dataset and source group combination. The detector signal yield
            instances are used to calculate the dataset signal weight factors.
            The order must follow the definition order of the log-likelihood
            ratio functions, i.e. datasets, and the definition order of the
            source hypothesis groups.
        """

        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        super(MultiSourceDatasetSignalWeights, self).__init__(
            src_hypo_group_manager, src_fitparam_mapper, detsigyields)

    def __call__(self, fitparam_values):
        """Calculates the dataset signal weight and its fit parameter gradients
        for each dataset.

        Parameters
        ----------
        fitparam_values : (N_fitparams+1,)-shaped 1D numpy ndarray
            The ndarray holding the current values of the fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        f : (N_datasets,)-shaped 1D ndarray
            The dataset signal weight factor for each dataset.
        f_grads : (N_datasets,N_fitparams)-shaped 2D ndarray | None
            The gradients of the dataset signal weight factors, one for each
            fit parameter. None is returned if there are no fit parameters
            beside ns.
        """
        fitparams_arr = self._src_fitparam_mapper.get_fitparams_array(fitparam_values[1:])

        N_datasets = self.n_datasets
        N_fitparams = self._src_fitparam_mapper.n_global_fitparams

        Y = np.empty((N_datasets, len(self._src_arr_list[0])), dtype=np.float64)
        if(N_fitparams > 0):
            Y_grads = np.empty((N_datasets, len(self._src_arr_list[0]), N_fitparams), dtype=np.float64)

        # Loop over the detector signal efficiency instances for the first and
        # only source hypothesis group.
        for (k, detsigyield_k) in enumerate(self._detsigyield_arr):
            for (j, detsigyield) in enumerate(detsigyield_k):
                (Yj, Yj_grads) = detsigyield(self._src_arr_list[k], fitparams_arr)
                # Store the detector signal yield and its fit parameter
                # gradients for the first and only source (element 0).
                Y[j] = Yj
                if(N_fitparams > 0):
                    Y_grads[j] = Yj_grads.T

        sum_Y = np.sum(Y)

        # f is a (N_datasets,)-shaped 1D ndarray.
        f = np.sum(Y, axis=1) / sum_Y

        # f_grads is a (N_datasets, N_fitparams)-shaped 2D ndarray.
        if(N_fitparams > 0):
            # sum_Y_grads is a (N_datasets, N_fitparams,)-shaped 2D array.
            sum_Y_grads = np.sum(Y_grads, axis=1)
            f_grads = (sum_Y_grads*sum_Y - (f*sum_Y)[...,np.newaxis]*np.sum(sum_Y_grads, axis=0)) / sum_Y**2
        else:
            f_grads = None

        return (f, f_grads)


class SourceWeights(object, metaclass=abc.ABCMeta):
    """Abstract base class for a source weight calculator class.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, detsigyields):
        """Constructs a new SourceWeights instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of the SourceHypoGroupManager managing the source
            hypothesis groups.
        src_fitparam_mapper : SourceFitParameterMapper
            The SourceFitParameterMapper instance that defines the global fit
            parameters and their mapping to the source fit parameters.
        detsigyields : (N_source_hypo_groups,)-shaped 1D ndarray of DetSigYield
                instances
            The collection of DetSigYield instances for each source hypothesis
            group.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self.src_fitparam_mapper = src_fitparam_mapper
        self.detsigyield_arr = np.atleast_1d(detsigyields)

        if(self._detsigyield_arr.shape[0] != self._src_hypo_group_manager.n_src_hypo_groups):
            raise ValueError('The detsigyields array must have the same number '
                'of source hypothesis groups as the source hypothesis group '
                'manager defines!')

        # Pre-convert the source list of each source hypothesis group into a
        # source array needed for the detector signal yield evaluation.
        # Since all the detector signal yield instances must be of the same
        # kind for each dataset, we can just use the one of the first dataset of
        # each source hypothesis group.
        self._src_arr_list = self._create_src_arr_list(
            self._src_hypo_group_manager, self._detsigyield_arr)

    def _create_src_arr_list(self, src_hypo_group_manager, detsigyield_arr):
        """Pre-convert the source list of each source hypothesis group into a
        source array needed for the detector signal yield evaluation.
        Since all the detector signal yield instances must be of the same
        kind for each dataset, we can just use the one of the first dataset of
        each source hypothesis group.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The SourceHypoGroupManager instance defining the sources.

        detsigyield_arr : (N_source_hypo_groups,)-shaped 1D ndarray of
                DetSigYield instances
            The collection of DetSigYield instances for each source hypothesis
            group.
        Returns
        -------
        src_arr_list : list of numpy record ndarrays
            The list of the source numpy record ndarrays, one for each source
            hypothesis group, which is needed by the detector signal yield
            instance.
        """
        src_arr_list = []
        for (gidx, src_hypo_group) in enumerate(src_hypo_group_manager.src_hypo_group_list):
            src_arr_list.append(
                detsigyield_arr[gidx].source_to_array(src_hypo_group.source_list)
            )

        return src_arr_list

    @property
    def src_hypo_group_manager(self):
        """The instance of SourceHypoGroupManager, which defines the source
        hypothesis groups.
        """
        return self._src_hypo_group_manager
    @src_hypo_group_manager.setter
    def src_hypo_group_manager(self, manager):
        if(not isinstance(manager, SourceHypoGroupManager)):
            raise TypeError('The src_hypo_group_manager property must be an '
                'instance of SourceHypoGroupManager!')
        self._src_hypo_group_manager = manager

    @property
    def src_fitparam_mapper(self):
        """The SourceFitParameterMapper instance defining the global fit
        parameters and their mapping to the source fit parameters.
        """
        return self._src_fitparam_mapper
    @src_fitparam_mapper.setter
    def src_fitparam_mapper(self, mapper):
        if(not isinstance(mapper, SourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper property must be an '
                'instance of SourceFitParameterMapper!')
        self._src_fitparam_mapper = mapper

    @property
    def detsigyield_arr(self):
        """The (N_source_hypo_groups,)-shaped 1D ndarray of DetSigYield
        instances.
        """
        return self._detsigyield_arr
    @detsigyield_arr.setter
    def detsigyield_arr(self, detsigyields):
        if(not isinstance(detsigyields, np.ndarray)):
            raise TypeError('The detsigyield_arr property must be an instance '
                'of numpy.ndarray!')
        if(detsigyields.ndim != 1):
            raise ValueError('The detsigyield_arr property must be a '
                'numpy.ndarray with 1 dimensions!')
        if(not issequenceof(detsigyields.flat, DetSigYield)):
            raise TypeError('The detsigyield_arr property must contain '
                'DetSigYield instances, one for each source hypothesis group!')
        self._detsigyield_arr = detsigyields

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the SourceHypoGroupManager instance of this
        DatasetSignalWeights instance. This will also recreate the internal
        source numpy record arrays needed for the detector signal efficiency
        calculation.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance, that should be used for
            this dataset signal weights instance.
        """
        self.src_hypo_group_manager = src_hypo_group_manager
        self._src_arr_list = self._create_src_arr_list(
            self._src_hypo_group_manager, self._detsigyield_arr)

    @abc.abstractmethod
    def __call__(self, fitparam_values):
        """This method is supposed to calculate source weights and
        their gradients.

        Parameters
        ----------
        fitparam_values : (N_fitparams+1,)-shaped 1D numpy ndarray
            The ndarray holding the current values of the fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        f : (N_sources,)-shaped 1D ndarray
            The source weight factor for each source.
        f_grads : (N_sources,)-shaped 1D ndarray | None
            The gradients of the source weight factors. None is returned if
            there are no fit parameters beside ns.
        """
        pass


class MultiPointSourcesRelSourceWeights(SourceWeights):
    """This class calculates the relative source weights for a group of point
    sources.
    """
    def __init__(
            self, src_hypo_group_manager, src_fitparam_mapper, detsigyields):
        """Constructs a new MultiPointSourcesRelSourceWeights instance assuming
        multiple sources.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The instance of the SourceHypoGroupManager managing the source
            hypothesis groups.
        src_fitparam_mapper : SingleSourceFitParameterMapper
            The instance of SingleSourceFitParameterMapper defining the global
            fit parameters and their mapping to the source fit parameters.
        detsigyields : (N_source_hypo_groups,)-shaped 1D ndarray of
                DetSigYield instances
            The collection of DetSigYield instances for each source hypothesis
            group.
        """

        if(not isinstance(src_fitparam_mapper, SingleSourceFitParameterMapper)):
            raise TypeError('The src_fitparam_mapper argument must be an '
                'instance of SingleSourceFitParameterMapper!')

        super(MultiPointSourcesRelSourceWeights, self).__init__(
            src_hypo_group_manager, src_fitparam_mapper, detsigyields)

    def __call__(self, fitparam_values):
        """Calculates the source weights and its fit parameter gradients
        for each source.

        Parameters
        ----------
        fitparam_values : (N_fitparams+1,)-shaped 1D numpy ndarray
            The ndarray holding the current values of the fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        f : (N_sources,)-shaped 1D ndarray
            The source weight factor for each source.
        f_grads : (N_sources,)-shaped 1D ndarray | None
            The gradients of the source weight factors. None is returned if
            there are no fit parameters beside ns.
        """
        fitparams_arr = self._src_fitparam_mapper.get_fitparams_array(fitparam_values[1:])

        N_fitparams = self._src_fitparam_mapper.n_global_fitparams

        Y = []
        Y_grads = []

        # Loop over detector signal efficiency instances for each source
        # hypothesis group in source hypothesis group manager.
        for (g, detsigyield) in enumerate(self._detsigyield_arr):
            (Yg, Yg_grads) = detsigyield(self._src_arr_list[g], fitparams_arr)

            # Store the detector signal yield and its fit parameter
            # gradients for all sources.
            Y.append(Yg)
            if(N_fitparams > 0):
                Y_grads.append(Yg_grads.T)

        Y = np.array(Y)
        sum_Y = np.sum(Y)

        # f is a (N_sources,)-shaped 1D ndarray.
        f = Y / sum_Y

        # Flatten the array so that each relative weight corresponds to specific
        # source.
        f = f.flatten()

        if(N_fitparams > 0):
            Y_grads = np.concatenate(Y_grads)

            # Sum over fit parameter gradients axis.
            # f_grads is a (N_sources,)-shaped 1D ndarray.
            f_grads = np.sum(Y_grads, axis=1) / sum_Y
        else:
            f_grads = None

        return (f, f_grads)


class MultiDatasetTCLLHRatio(TCLLHRatio):
    """This class describes a two-component log-likelihood ratio function for
    multiple datasets. The final log-likelihood ratio value is the sum of the
    individual log-likelihood ratio values.

    The different datasets contribute according to their dataset signal weight
    factor, f_j(p_s), which depends on possible signal fit parameters. By
    definition the signal fit parameters are assumed to be the same for each
    dataset.

    By mathematical definition this class is suitable for single and multi
    source hypotheses.
    """
    def __init__(self, minimizer, dataset_signal_weights, llhratios):
        """Creates a new composite two-component log-likelihood ratio function.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        dataset_signal_weights : DatasetSignalWeights
            An instance of DatasetSignalWeights, which calculates the relative
            dataset weight factors.
        llhratios : sequence of SingleDatasetTCLLHRatio instances
            The sequence of the two-component log-likelihood ratio functions,
            one for each dataset.
        """
        self.dataset_signal_weights = dataset_signal_weights
        self.llhratio_list = llhratios

        super(MultiDatasetTCLLHRatio, self).__init__(
            minimizer, self._llhratio_list[0].mean_n_sig_0)

        # Check if the number of datasets the DatasetSignalWeights instance is
        # made for equals the number of log-likelihood ratio functions.
        if(self.dataset_signal_weights.n_datasets != len(self._llhratio_list)):
            raise ValueError('The number of datasets the DatasetSignalWeights '
                'instance is made for must be equal to the number of '
                'log-likelihood ratio functions!')

        # Define cache variable for the dataset signal weight factors, which
        # will be needed when calculating the second derivative w.r.t. ns of the
        # log-likelihood ratio function.
        self._cache_fitparam_values_ns = None
        self._cache_f = None

    @property
    def dataset_signal_weights(self):
        """The DatasetSignalWeights instance that provides the relative dataset
        weight factors.
        """
        return self._dataset_signal_weights
    @dataset_signal_weights.setter
    def dataset_signal_weights(self, obj):
        if(not isinstance(obj, DatasetSignalWeights)):
            raise TypeError('The dataset_signal_weights property must be an '
                'instance of DatasetSignalWeights!')
        self._dataset_signal_weights = obj

    @property
    def llhratio_list(self):
        """(read-only) The list of TCLLHRatio instances, which are part of this
        composite log-likelihood-ratio function.
        """
        return self._llhratio_list
    @llhratio_list.setter
    def llhratio_list(self, llhratios):
        if(not issequenceof(llhratios, SingleDatasetTCLLHRatio)):
            raise TypeError('The llhratio_list property must be a sequence of '
                'SingleDatasetTCLLHRatio instances!')
        self._llhratio_list = list(llhratios)

    @property
    def n_selected_events(self):
        """(read-only) The sum of selected events of each individual
        log-likelihood ratio function.
        """
        n_selected_events = 0
        for llhratio in self._llhratio_list:
            n_selected_events += llhratio.tdm.n_selected_events
        return n_selected_events

    @TCLLHRatio.mean_n_sig_0.setter
    def mean_n_sig_0(self, v):
        TCLLHRatio.mean_n_sig_0.fset(self, v)
        for llhratio in self._llhratio_list:
            llhratio.mean_n_sig_0 = self._mean_n_sig_0

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the source hypo group manager of all objects of this LLH
        ratio function, hence, calling the `change_source_hypo_group_manager`
        method of all TCLLHRatio objects of this LLHRatio instance.
        """
        # Change the source hypo group manager of the DatasetSignalWeights
        # instance.
        self._dataset_signal_weights.change_source_hypo_group_manager(
            src_hypo_group_manager)

        for llhratio in self._llhratio_list:
            llhratio.change_source_hypo_group_manager(src_hypo_group_manager)

    def initialize_for_new_trial(self, tl=None):
        """Initializes the log-likelihood-ratio function for a new trial.
        """
        for llhratio in self._llhratio_list:
            llhratio.initialize_for_new_trial(tl=tl)

    def evaluate(self, fitparam_values, tl=None):
        """Evaluates the composite log-likelihood-ratio function and returns its
        value and global fit parameter gradients.

        Parameters
        ----------
        fitparam_values : (N_fitparams)-shaped numpy 1D ndarray
            The ndarray holding the current values of the global fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value of the composite
            log-likelihood-ratio function.
        grads : (N_fitparams,)-shaped 1D ndarray
            The ndarray holding the gradient value of the composite
            log-likelihood-ratio function for ns and each global fit parameter.
            By definition the first element is the gradient for ns.
        """
        tracing = CFG['debugging']['enable_tracing']

        ns = fitparam_values[0]
        if(tracing):
            logger.debug(
                '{:s}.evaluate: ns={:.3f}'.format(
                    classname(self), ns))

        # Get the dataset signal weights and their gradients.
        # f is a (N_datasets,)-shaped 1D ndarray.
        # f_grads is a (N_datasets,N_fitparams)-shaped 2D ndarray.
        (f, f_grads) = self._dataset_signal_weights(fitparam_values)

        # Cache f for possible later calculation of the second derivative w.r.t.
        # ns of the log-likelihood ratio function.
        self._cache_fitparam_values_ns = ns
        self._cache_f = f

        nsf = ns * f

        # Calculate the composite log-likelihood-ratio function value and the
        # gradient of the composite log-likelihood ratio function for each
        # global fit parameter.
        log_lambda = 0

        # Allocate an array for the gradients of the composite log-likelihood
        # function. It is always at least one element long, i.e. the gradient
        # for ns.
        grads = np.zeros((len(fitparam_values),), dtype=np.float64)

        # Create an array holding the fit parameter values for a particular
        # llh ratio function. Since we need to adjust ns with nsj it's more
        # efficient to create this array once and use it within the for loop
        # over the llh ratio functions.
        llhratio_fitparam_values = np.empty(
            (len(fitparam_values),), dtype=np.float64)
        # Loop over the llh ratio functions.
        for (j, llhratio) in enumerate(self._llhratio_list):
            if(tracing):
                logger.debug(
                    'nsf[j={:d}] = {:.3f}'.format(
                        j, nsf[j]))
            llhratio_fitparam_values[0] = nsf[j]
            llhratio_fitparam_values[1:] = fitparam_values[1:]
            (log_lambda_j, grads_j) = llhratio.evaluate(
                llhratio_fitparam_values, tl=tl)
            log_lambda += log_lambda_j
            # Gradient for ns.
            grads[0] += grads_j[0] * f[j]
            # Gradient for each global fit parameter, if there are any.
            if(len(grads) > 1):
                grads[1:] += grads_j[0] * ns * f_grads[j] + grads_j[1:]

        return (log_lambda, grads)

    def calculate_ns_grad2(self, fitparam_values):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.
        This method tries to use cached values for the dataset signal weight
        factors. If cached values don't exist or do not match the given fit
        parameter values, they will get calculated automatically by calling the
        evaluate method with the given fit parameter values.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        ns = fitparam_values[0]

        # Check if the cached fit parameters match the given ones. The ns value
        # is special to the multi-dataset LLH ratio function, but all the other
        # fit parameters are shared by all the LLH ratio functions of the
        # different datasets. So those we just query from the first LLH ratio
        # function.
        if((self._cache_fitparam_values_ns is None) or
           (self._cache_fitparam_values_ns != ns) or
           (not np.all(self._llhratio_list[0]._cache_fitparam_values[1:] == fitparam_values[1:]))):
            self.evaluate(fitparam_values)

        nsf = ns * self._cache_f

        nsgrad2j = np.empty((len(self._llhratio_list),), dtype=np.float64)
        # Loop over the llh ratio functions and their second derivative.
        llhratio_fitparam_values = np.empty(
            (len(fitparam_values),), dtype=np.float64)
        for (j, llhratio) in enumerate(self._llhratio_list):
            llhratio_fitparam_values[0] = nsf[j]
            llhratio_fitparam_values[1:] = fitparam_values[1:]
            nsgrad2j[j] = llhratio.calculate_ns_grad2(llhratio_fitparam_values)

        nsgrad2 = np.sum(nsgrad2j * self._cache_f**2)

        return nsgrad2


class NsProfileMultiDatasetTCLLHRatio(TCLLHRatio):
    r"""This class implements a profile log-likelihood ratio function that has
    only ns as fit parameter. It uses a MultiDatasetTCLLHRatio instance as
    log-likelihood function. Hence, mathematically it is

    .. math::

        \Lambda(n_s) = \frac{L(n_s)}{L(n_s=n_{s,0})},

    where :math:`n_{s,0}` is the fixed mean number of signal events for the
    null-hypothesis.
    """
    def __init__(self, minimizer, mean_n_sig_0, llhratio):
        """Creates a new ns-profile log-likelihood-ratio function with a
        null-hypothesis where ns is fixed to `mean_n_sig_0`.

        Parameters
        ----------
        minimizer : instance of Minimizer
            The Minimizer instance that should be used to minimize the negative
            of this log-likelihood ratio function.
        mean_n_sig_0 : float
            The fixed parameter value for the mean number of signal events of
            the null-hypothesis.
        llhratio : instance of MultiDatasetTCLLHRatio
            The instance of MultiDatasetTCLLHRatio, which should be used as
            log-likelihood function.
        """
        super(NsProfileMultiDatasetTCLLHRatio, self).__init__(
            minimizer, mean_n_sig_0)

        self.llhratio = llhratio

        # Check that the given log-likelihood-ratio function has no fit
        # parameters, i.e. only ns in the end.
        for sub_llhratio in llhratio.llhratio_list:
            n_global_fitparams = sub_llhratio.src_fitparam_mapper.n_global_fitparams
            if(n_global_fitparams != 0):
                raise ValueError('The log-likelihood-ratio functions of the '
                    'MultiDatasetTCLLHRatio instance must have no global fit '
                    'parameters, i.e. only ns in the end! Currently it has %d '
                    'global fit parameters'%(n_global_fitparams))

        # Define a member to hold the constant null-hypothesis log-likelihood
        # function value for ns=mean_n_sig_0.
        self._logL_0 = None

    @property
    def llhratio(self):
        """The instance of MultiDatasetTCLLHRatio, which should be used as
        log-likelihood function.
        """
        return self._llhratio
    @llhratio.setter
    def llhratio(self, obj):
        if(not isinstance(obj, MultiDatasetTCLLHRatio)):
            raise TypeError('The llhratio property must be an instance of '
                'MultiDatasetTCLLHRatio!')
        self._llhratio = obj

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the source hypo group manager of all objects of this LLH
        ratio function, hence, calling the `change_source_hypo_group_manager`
        method of the underlaying MultiDatasetTCLLHRatio instance of this
        LLHRatio instance.
        """
        self._llhratio.change_source_hypo_group_manager(src_hypo_group_manager)

    def initialize_for_new_trial(self, tl=None):
        """Initializes the log-likelihood-ratio function for a new trial.

        Parameters
        ----------
        tl : TimeLord | None
            The optional TimeLord instance that should be used for timing
            measurements.
        """
        self._llhratio.initialize_for_new_trial(tl=tl)

        # Compute the constant log-likelihood function value for the
        # null-hypothesis.
        fitparam_values_0 = np.array([self._mean_n_sig_0], dtype=np.float64)
        (self._logL_0, grads_0) = self._llhratio.evaluate(fitparam_values_0)

    def evaluate(self, fitparam_values):
        """Evaluates the log-likelihood-ratio function and returns its value and
        global fit parameter gradients.

        Parameters
        ----------
        fitparam_values : (N_fitparams)-shaped numpy 1D ndarray
            The ndarray holding the current values of the global fit parameters.
            The first element of that array is, by definition, the number of
            signal events, ns.

        Returns
        -------
        log_lambda : float
            The calculated log-lambda value of this log-likelihood-ratio
            function.
        grads : (N_fitparams,)-shaped 1D ndarray
            The ndarray holding the gradient value of this log-likelihood-ratio
            for ns.
            By definition the first element is the gradient for ns.
        """
        (logL, grads) = self._llhratio.evaluate(fitparam_values)

        return (logL - self._logL_0, grads)

    def calculate_ns_grad2(self, fitparam_values):
        """Calculates the second derivative w.r.t. ns of the log-likelihood
        ratio function.

        Parameters
        ----------
        fitparam_values : numpy (N_fitparams+1)-shaped 1D ndarray
            The ndarray holding the current values of the fit parameters.
            By definition, the first element is the fit parameter for the number
            of signal events, ns.

        Returns
        -------
        nsgrad2 : float
            The second derivative w.r.t. ns of the log-likelihood ratio function
            for the given fit parameter values.
        """
        return self._llhratio.calculate_ns_grad2(fitparam_values)

#class NestedProfileLLHRatio(LLHRatio, metaclass=abc.ABCMeta):
    #r"""This class provides the abstract base class for a nested profile
    #log-likelihood ratio function, which is, by definition, of the form

    #.. math::

        #\Lambda = \frac{\sup_{\Theta_0} L(\theta|D)}{\sup_{\Theta} L(\theta|D)}

    #where :math:`\theta` are the possible fit parameters, and :math:`\Theta`
    #and :math:`\Theta_0` are the total and nested fit parameter spaces,
    #respectively.
    #"""

    #def __init__(self, ):
        #super(NestedProfileLLHRatio, self).__init__()


#class MultiDatasetNestedProfileLLHRatio(NestedProfileLLHRatio):
    #"""This class provides a nested profile log-likelihood ratio function for
    #multiple data sets.
    #"""
    #def __init__(self):
        #super(MultiDatasetNestedProfileLLHRatio, self).__init__()




# -*- coding: utf-8 -*-

"""The ``backgroundpdf`` module contains possible background PDF models for the
likelihood function.
"""

from skyllh.core.pdf import (
    IsBackgroundPDF,
    MultiDimGridPDF,
    NDPhotosplinePDF,
    TimePDF,
)

import numpy as np


class BackgroundMultiDimGridPDF(MultiDimGridPDF, IsBackgroundPDF):
    """This class provides a multi-dimensional background PDF. The PDF is
    created from pre-calculated PDF data on a grid. The grid data is
    interpolated using a :class:`scipy.interpolate.RegularGridInterpolator`
    instance.
    """

    def __init__(
            self,
            axis_binnings,
            path_to_pdf_splinetable=None,
            pdf_grid_data=None,
            norm_factor_func=None):
        """Creates a new background PDF instance for a multi-dimensional PDF
        given as PDF values on a grid. The grid data is interpolated with a
        :class:`scipy.interpolate.RegularGridInterpolator` instance. As grid
        points the bin edges of the axis binning definitions are used.

        Parameters
        ----------
        axis_binnings : sequence of BinningDefinition
            The sequence of BinningDefinition instances defining the binning of
            the PDF axes. The name of each BinningDefinition instance defines
            the event field name that should be used for querying the PDF.
        path_to_pdf_splinetable : str
            The path to the file containing the spline table.
            The spline table contains a pre-computed fit to pdf_grid_data.
        pdf_grid_data : n-dimensional numpy ndarray
            The n-dimensional numpy ndarray holding the PDF values at given grid
            points. The grid points must match the bin edges of the given
            BinningDefinition instances of the `axis_binnings` argument.
        norm_factor_func : callable | None
            The function that calculates a possible required normalization
            factor for the PDF value based on the event properties.
            The call signature of this function
            must be `__call__(pdf, events, fitparams)`, where `pdf` is this PDF
            instance, `events` is a numpy record ndarray holding the events for
            which to calculate the PDF values, and `fitparams` is a dictionary
            with the current fit parameter names and values.
        """
        super(BackgroundMultiDimGridPDF, self).__init__(
            axis_binnings, path_to_pdf_splinetable, pdf_grid_data, norm_factor_func)


class BackgroundNDPhotosplinePDF(NDPhotosplinePDF, IsBackgroundPDF):
    """This class provides a multi-dimensional background PDF created from a
    n-dimensional photospline fit. The photospline package is used to evaluate
    the PDF fit.
    """

    def __init__(
            self,
            axis_binnings,
            param_set,
            path_to_pdf_splinefit,
            norm_factor_func=None):
        """Creates a new background PDF instance for a n-dimensional photospline
        PDF fit.

        Parameters
        ----------
        axis_binnings : BinningDefinition | sequence of BinningDefinition
            The sequence of BinningDefinition instances defining the binning of
            the PDF axes. The name of each BinningDefinition instance defines
            the event field name that should be used for querying the PDF.
        param_set : Parameter | ParameterSet
            The Parameter instance or ParameterSet instance defining the
            parameters of this PDF. The ParameterSet holds the information
            which parameters are fixed and which are floating (i.e. fitted).
        path_to_pdf_splinefit : str
            The path to the file containing the photospline fit.
        norm_factor_func : callable | None
            The function that calculates a possible required normalization
            factor for the PDF value based on the event properties.
            The call signature of this function must be
            `__call__(pdf, tdm, params)`, where `pdf` is this PDF
            instance, `tdm` is an instance of TrialDataManager holding the
            event data for which to calculate the PDF values, and `params` is a
            dictionary with the current parameter names and values.
        """
        super(BackgroundNDPhotosplinePDF, self).__init__(
            axis_binnings=axis_binnings,
            param_set=param_set,
            path_to_pdf_splinefit=path_to_pdf_splinefit,
            norm_factor_func=norm_factor_func
        )


class BackgroundUniformTimePDF(TimePDF, IsBackgroundPDF):

    def __init__(self, grl):
        """Creates a new background time PDF instance as uniform background

        Parameters
        ----------
        grl : ndarray
            Array of the detector good run list

        """
        super(BackgroundUniformTimePDF, self).__init__()
        self.start = grl["start"][0]
        self.end = grl["stop"][-1]
        self.grl = grl


    def cdf(self, t):
        """Compute the cumulative density function for the box pdf. This is
        needed for normalization.

        Parameters
        ----------
        t : float, ndarray
            MJD times

        Returns
        -------
        cdf : float, ndarray
            Values of cumulative density function evaluated at t
        """
        t_start = self.grl["start"][0]
        t_end = self.grl["stop"][-1]
        t = np.atleast_1d(t)

        cdf = np.zeros(t.size, float)

        # values between start and stop times
        mask = (t_start <= t) & (t <= t_end)
        cdf[mask] = (t[mask] - t_start) / [t_end - t_start]

        # take care of values beyond stop time in sample

        return cdf

    def norm_uptime(self):
        """Compute the normalization with the dataset uptime. Distributions like 
        scipy.stats.norm are normalized (-inf, inf).
        These must be re-normalized such that the function sums to 1 over the
        finite good run list domain.

        Returns
        -------
        norm : float
            Normalization such that cdf sums to 1 over good run list domain
        """

        integral = (self.cdf(self.grl["stop"]) - self.cdf(self.grl["start"])).sum()

        if np.isclose(integral, 0):
            return 0

        return 1. / integral

    def get_prob(self, tdm, fitparams=None, tl=None):
        """Calculates the background time probability density of each event.

        tdm : TrialDataManager
            Unused interface argument.
        fitparams : None
            Unused interface argument.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        pd : array of float
            The (N,)-shaped ndarray holding the probability density for each event.
        grads : empty array of float
            Does not depend on fit parameter, so no gradient.
        """
        livetime = self.grl["stop"][-1] - self.grl["start"][0]
        pd = 1./livetime
        grads = np.array([], dtype=np.double)

        return (pd, grads)



# -*- coding: utf-8 -*-

from skyllh.core.background_generation import BackgroundGenerationMethod
from skyllh.core.dataset import Dataset, DatasetData
from skyllh.core.py import issequenceof
from skyllh.core.random import RandomStateService


class BackgroundGenerator(object):
    """This is the general background generator class, which provides a method
    to generate background events. It does not depend on the detector or
    background hypothesis. These dependencies are out-sourced to a class derived
    from the BackgroundGenerationMethod class.
    """
    def __init__(self, bkg_gen_method, dataset_list, data_list):
        """Constructs a new background generator instance.

        Parameters
        ----------
        bkg_gen_method : instance of BackgroundGenerationMethod
            The background event generation method, which should be used to
            generate events.
        dataset_list : list of Dataset instances
            The list of Dataset instances for which background events should get
            generated for.
        data_list : list of DatasetData instances
            The list of DatasetData instances holding the actual data of each
            dataset. The order must match the order of ``dataset_list``.
        """
        super(BackgroundGenerator, self).__init__()

        self.bkg_gen_method = bkg_gen_method
        self.dataset_list = dataset_list
        self.data_list = data_list

    @property
    def bkg_gen_method(self):
        """The instance of BackgroundGenerationMethod which should be used to
        generate background events.
        """
        return self._bkg_gen_method
    @bkg_gen_method.setter
    def bkg_gen_method(self, method):
        if(not isinstance(method, BackgroundGenerationMethod)):
            raise TypeError('The bkg_gen_method property must be an instance '
                'of BackgroundGenerationMethod!')
        self._bkg_gen_method = method

    @property
    def dataset_list(self):
        """The list of Dataset instances for which background events should get
        generated for.
        """
        return self._dataset_list
    @dataset_list.setter
    def dataset_list(self, datasets):
        if(not issequenceof(datasets, Dataset)):
            raise TypeError('The dataset_list property must be a sequence of '
                'Dataset instances!')
        self._dataset_list = list(datasets)

    @property
    def data_list(self):
        """The list of DatasetData instances holding the actual data of each
        dataset. The order must match the order of the ``dataset_list``
        property.
        """
        return self._data_list
    @data_list.setter
    def data_list(self, datas):
        if(not issequenceof(datas, DatasetData)):
            raise TypeError('The data_list property must be a sequence of '
                'DatasetData instances!')
        self._data_list = datas

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the SourceHypoGroupManager instance of the background
        generation method.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The new SourceHypoGroupManager instance.
        """
        self._bkg_gen_method.change_source_hypo_group_manager(
            src_hypo_group_manager)

    def generate_background_events(self, rss, dataset_idx, tl=None, **kwargs):
        """Generates a mean number of background events for the given dataset.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset_idx : int
            The index of the dataset for which background events should get
            generated for.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which will be passed to the
            ``generate_events`` method of the background generation method
            instance. Usual keyword arguments are `mean`, the mean number of
            background events to generate, and `poisson`, the flag if the number
            of background events should get drawn from a Poisson distribution
            with the given mean number of background events as mean.

        Returns
        -------
        n_bkg : int
            The number of generated background events.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events can be less than `n_bkg`
            if an event selection method is used.
        """
        ds = self._dataset_list[dataset_idx]
        data = self._data_list[dataset_idx]

        (n_bkg, bkg_events) = self._bkg_gen_method.generate_events(
            rss, ds, data, tl=tl, **kwargs)

        return (n_bkg, bkg_events)



# -*- coding: utf-8 -*-

from skyllh.core.background_generation import BackgroundGenerationMethod
from skyllh.core.scrambling import DataScrambler


class FixedScrambledExpDataI3BkgGenMethod(BackgroundGenerationMethod):
    """This class implements the background event generation method for the
    IceCube detector using scrambled experimental data as background hypothesis
    with a fixed number of background events equal to the number of events in
    the dataset. This background generation method is the one used in SkyLab.
    """
    def __init__(self, data_scrambler):
        """Creates a new background generation method instance to generate
        background events from scrambled experimental data with a fixed number
        of events equal to the number of events in the dataset.

        Parameters
        ----------
        data_scrambler : instance of DataScrambler
            The DataScrambler instance to use to generate scrambled experimental
            data.
        """
        super(FixedScrambledExpDataI3BkgGenMethod, self).__init__()

        self.data_scrambler = data_scrambler

    @property
    def data_scrambler(self):
        """The DataScrambler instance that implements the data scrambling.
        """
        return self._data_scrambler
    @data_scrambler.setter
    def data_scrambler(self, scrambler):
        if(not isinstance(scrambler, DataScrambler)):
            raise TypeError('The data_scrambler property must be an instance '
                'of DataScrambler!')
        self._data_scrambler = scrambler

    def generate_events(self, rss, dataset, data, **kwargs):
        """Generates background events from the given data, by scrambling the
        data. The number of events is equal to the size of the given dataset.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from. It is used to scramble the experimental data.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.

        Returns
        -------
        n_bkg : int
            The number of generated background events.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events.
        """
        # Scramble the experimental data events, but make a copy first.
        bkg_events = self._data_scrambler.scramble_data(rss, data.exp.copy())

        return (len(bkg_events), bkg_events)



# -*- coding: utf-8 -*-

"""This module defines the base classes for any physics models used by SkyLLH.
"""

from skyllh.core.py import (
    ObjectCollection,
    issequence,
    str_cast
)


class PhysicsModel(object):
    """This class provides a base class for all physics models like source
    models or background models.
    """
    def __init__(self, name=None):
        """Creates a new PhysicsModel instance.

        Parameters
        ----------
        name : str | None
            The name of the physics model. If set to `None`, the id of the
            object is taken as name.
        """
        super(PhysicsModel, self).__init__()

        if(name is None):
            name = self.id

        self.name = name

    @property
    def name(self):
        """The name of the physics model.
        """
        return self._name
    @name.setter
    def name(self, n):
        n = str_cast(n, 'The name property must be castable to type str!')
        self._name = n

    @property
    def id(self):
        """(read-only) The ID of the physics model. It's an integer generated
        with Python's `id` function. Hence, it's related to the memory address
        of the object.
        """
        return id(self)


class PhysicsModelCollection(ObjectCollection):
    """This class describes a collection of PhysicsModel instances. It can be
    used to group several physics models into a single object.
    """
    @staticmethod
    def cast(obj, errmsg):
        """Casts the given object to a PhysicsModelCollection object.
        If the cast fails, a TypeError with the given error message is raised.

        Parameters
        ----------
        obj : PhysicsModel instance | sequence of PhysicsModel instances |
                PhysicsModelCollection | None
            The object that should be casted to PhysicsModelCollection.
        errmsg : str
            The error message if the cast fails.

        Raises
        ------
        TypeError
            If the cast fails.

        Returns
        -------
        physmodelcollection : instance of PhysicsModelCollection
            The created PhysicsModelCollection instance. If `obj` is already a
            PhysicsModelCollection instance, it will be returned.
        """
        if(obj is None):
            obj = PhysicsModelCollection(PhysicsModel, None)
            return obj

        if(isinstance(obj, PhysicsModel)):
            obj = PhysicsModelCollection(PhysicsModel, [obj])
            return obj

        if(isinstance(obj, PhysicsModelCollection)):
            return obj

        if(issequence(obj)):
            obj = PhysicsModelCollection(PhysicsModel, obj)
            return obj

        raise TypeError(errmsg)

    def __init__(self, model_type=None, models=None):
        """Creates a new PhysicsModel collection. The type of the physics model
        instances the collection holds can be restricted, by setting the
        model_type parameter.

        Parameters
        ----------
        model_type : type | None
            The type of the physics model. If set to None (default),
            PhysicsModel will be used.
        models : sequence of model_type instances | None
            The sequence of physics models this collection should be initalized
            with.
        """
        if(model_type is None):
            model_type = PhysicsModel
        super(PhysicsModelCollection, self).__init__(
            obj_type=model_type,
            obj_list=models)

    @property
    def model_type(self):
        """(read-only) The type of the physics model.
        """
        return self.obj_type

    @property
    def models(self):
        """(read-only) The list of models of type `model_type`.
        """
        return self.objects



# -*- coding: utf-8 -*-

"""Note: This module is deprecated and new flux models should be implemented in
    `flux_model.py`. However, the framework currently doesn't support flux
    models derived from `flux_model.FluxModel`.

The ``flux`` module contains all standard flux models for a source.
The abstract class ``FluxModel`` serves as a base class for all flux model
classes.
The unit of the resulting flux value must be [energy]^-1 [length]^-2 [time]^-1.
The units are defined using the astropy.units module and can be set through
the properties ``energy_unit``, ``length_unit``, and ``time_unit``.
The default units are [energy] = GeV, [length] = cm, [time] = s.
"""
import abc
import numpy as np

from copy import deepcopy

from astropy import units

from skyllh.core.py import classname, isproperty, float_cast
from skyllh.core.config import CFG


def get_conversion_factor_to_internal_flux_unit(fluxmodel):
    """Calculates the unit conversion factor for converting the used flux
    unit of the given flux model into the skyllh internally used flux unit
    1/(GeV cm2 s).

    Parameters
    ----------
    fluxmodel : FluxModel
        The flux model instance for which to calculate the unit conversion
        factor.

    Returns
    -------
    unit_conversion_factor : float
        The unit conversion factor.
    """
    fluxmodel_flux_unit = 1/(
        fluxmodel.energy_unit * fluxmodel.length_unit**2 * fluxmodel.time_unit)

    internal_units = CFG['internal_units']
    internal_flux_unit = 1/(
        internal_units['energy'] * internal_units['length']**2 *
        internal_units['time'])

    unit_conversion_factor = (fluxmodel_flux_unit).to(internal_flux_unit).value
    return unit_conversion_factor


class FluxModel(object, metaclass=abc.ABCMeta):
    """Abstract base class for all flux models.
    This base class defines the units used for the flux calculation. At this
    point the function form of the flux model is not yet defined.

    Attributes
    ----------
    energy_unit : str
        The used unit of energy.
    length_unit : str
        The used unit of length.
    time_unit : str
        The used unit of time.
    math_function_str : str
        The string showing the mathematical function of the flux calculation.
    """

    def __init__(self):
        super(FluxModel, self).__init__()

        # Define the default units.
        self.energy_unit = units.GeV
        self.length_unit = units.cm
        self.time_unit = units.s

    @property
    def energy_unit(self):
        """The unit of energy used for the flux calculation.
        """
        return self._energy_unit
    @energy_unit.setter
    def energy_unit(self, unit):
        if(not isinstance(unit, units.UnitBase)):
            raise TypeError('The property energy_unit must be of type astropy.units.UnitBase!')
        self._energy_unit = unit

    @property
    def length_unit(self):
        """The unit of length used for the flux calculation.
        """
        return self._length_unit
    @length_unit.setter
    def length_unit(self, unit):
        if(not isinstance(unit, units.UnitBase)):
            raise TypeError('The property length_unit must be of type astropy.units.UnitBase!')
        self._length_unit = unit

    @property
    def time_unit(self):
        """The unit of length used for the flux calculation.
        """
        return self._time_unit
    @time_unit.setter
    def time_unit(self, unit):
        if(not isinstance(unit, units.UnitBase)):
            raise TypeError('The property time_unit must be of type astropy.units.UnitBase!')
        self._time_unit = unit

    @property
    def unit_str(self):
        """The string representation of the flux unit.
        """
        return '1/(%s %s^2 %s)'%(
            self.energy_unit.to_string(), self.length_unit.to_string(),
            self.time_unit.to_string())

    @property
    def unit_latex_str(self):
        """The latex string representation of the flux unit.
        """
        return r'%s$^{-1}$ %s$^{-2}$ %s$^{-1}$'%(
            self.energy_unit.to_string(), self.length_unit.to_string(),
            self.time_unit.to_string())

    @property
    @abc.abstractmethod
    def math_function_str(self):
        """The string showing the mathematical function of the flux calculation.
        """
        pass

    @abc.abstractmethod
    def __call__(self, E):
        """The call operator to retrieve a flux value for a given energy.

        Parameters
        ----------
        E : float | 1d numpy.ndarray of float
            The energy for which to retrieve the flux value.

        Returns
        -------
        flux : ndarray of float
            Flux at energy E in unit [energy]^-1 [length]^-2 [time]^-1.
            By default that is GeV^-1 cm^-2 s^-1.
        """
        pass

    def __str__(self):
        """Pretty string representation of this class.
        """
        return self.math_function_str + ' ' + self.unit_str

    def copy(self, newprop=None):
        """Copies this flux model object by calling the copy.deepcopy function,
        and sets new properties if requested.

        Parameters
        ----------
        newprop : dict | None
            The dictionary with the new property values to set, where the
            dictionary key is the property name and the dictionary value is the
            new value of the property.
        """
        fluxmodel = deepcopy(self)

        # Set the new property values.
        if(newprop is not None):
            fluxmodel.set_properties(newprop)

        return fluxmodel

    def set_properties(self, propdict):
        """Sets the properties of the flux model to the given property values.

        Parameters
        ----------
        propdict : dict (name: value)
            The dictionary holding the names of the properties and their new
            values.
        """
        if(not isinstance(propdict, dict)):
            raise TypeError('The propdict argument must be of type dict!')
        for (prop, val) in propdict.items():
            if(not hasattr(self, prop)):
                raise KeyError('The flux model "%s" does not have a property named "%s"!'%(classname(self), prop))
            if(not isproperty(self, prop)):
                raise TypeError('The attribute "%s" of flux model "%s" is no property!'%(classname(self), prop))
            setattr(self, prop, val)


class NormedFluxModel(FluxModel, metaclass=abc.ABCMeta):
    """Abstract base class for all normalized flux models of the form

        dN/(dEdAdt) = Phi0 * f(E/E0),

    where Phi0 is the flux normalization at E=E0 in the flux unit
    [energy]^-1 [length]^-2 [time]^-1, and f(E/E0) is the unit-less energy
    dependence of the flux.

    The unit of dN/(dEdAdt) is [energy]^-1 [length]^-2 [time]^-1.
    By default the unit is GeV^-1 cm^-2 s^-1.

    Attributes
    ----------
    Phi0 : float
        Flux value (dN/(dEdAdt)) at E0 in unit
        [energy]^-1 [length]^-2 [time]^-1.
    E0 : float
        Normalization energy in unit of energy.
    """

    def __init__(self, Phi0, E0):
        super(NormedFluxModel, self).__init__()

        self.Phi0 = Phi0
        self.E0 = E0

    @property
    def Phi0(self):
        """The flux value (dN/(dEdAdt)) at energy E0 in unit
        [energy]^-1 [length]^-2 [time]^-1.
        """
        return self._Phi0

    @Phi0.setter
    def Phi0(self, v):
        v = float_cast(v, 'Property Phi0 must be castable to type float!')
        self._Phi0 = v

    @property
    def E0(self):
        """The normalization energy.
        """
        return self._E0
    @E0.setter
    def E0(self, v):
        v = float_cast(v, 'Property E0 must be castable to type float!')
        self._E0 = v


class SplineFluxModel(FluxModel, metaclass=abc.ABCMeta):
    """Abstract base class for all flux models that are represented
    numerically through photospline splinetables in .fits format

        dN/(dEdAdt) = Phi0 * f(E),

    where Phi0 is the relative flux normalization (dimensionless)
    and f(E) represents the photospline in units of
    [energy]^-1 [length]^-2 [time]^-1 i.e. the energy dependence of the flux.

    The unit of dN/(dEdAdt) is [energy]^-1 [length]^-2 [time]^-1.
    By default the unit is GeV^-1 cm^-2 s^-1.

    Outside of support [crit_log_nu_energy_lower, crit_log_nu_energy_upper] the
    flux will be set to 0.

    Attributes
    ----------
    Phi0 : float
        Flux normalization relative to model prediction.
    psp_table : object
        The photospline.SplineTable object.
    crit_log_nu_energy_lower : float
        Lower end of energy range (support) of spline flux.
    crit_log_nu_energy_upper : float
        Upper end of energy range (support) of spline flux.
    """
    def __init__(self, Phi0, psp_table, crit_log_nu_energy_lower, crit_log_nu_energy_upper):
        super(SplineFluxModel, self).__init__()
        self._psp_table = psp_table
        self._Phi0 = Phi0
        self._crit_log_nu_energy_lower = crit_log_nu_energy_lower
        self._crit_log_nu_energy_upper = crit_log_nu_energy_upper

    @property
    def psp_table(self):
        """The photospline.SplineTable object that describes the neutrino flux
        as function of neutrino energy via B-spline interpolation.
        """
        return self._psp_table
    @psp_table.setter
    def psp_table(self, t):
        self._psp_table = t

    @property
    def Phi0(self):
        """The relative flux normalization. Phi0=1 corresponds to the nominal
        model flux.
        """
        return self._Phi0
    @Phi0.setter
    def Phi0(self, v):
        v = float_cast(v, 'Property Phi0 must be castable to type float!')
        self._Phi0 = v

    @property
    def crit_log_nu_energy_lower(self):
        """The lower bound of the support of the spline interpolator.
        """
        return self._crit_log_nu_energy_lower
    @crit_log_nu_energy_lower.setter
    def crit_log_nu_energy_lower(self, v):
        v = float_cast(
            v, 'Property crit_log_nu_energy_lower must be castable to type float!')
        self._crit_log_nu_energy_lower = v

    @property
    def crit_log_nu_energy_upper(self):
        """The upper bound of the support of the spline interpolator.
        """
        return self._crit_log_nu_energy_upper
    @crit_log_nu_energy_upper.setter
    def crit_log_nu_energy_upper(self, v):
        v = float_cast(
            v, 'Property crit_log_nu_energy_upper must be castable to type float!')
        self._crit_log_nu_energy_upper = v


class SeyfertCoreCoronaFlux(SplineFluxModel):
    """Implements the Core-Corona Seyfert Galaxy neutrino flux model of
    A. Kheirandish et al., Astrophys.J. 922 (2021) 45 by means of B-spline
    interpolation.

    Attributes
    ----------
    Phi0 : float
        Flux normalization relative to model prediction.
    log_xray_lumin : float
        log10 of intrinsic x-ray luminosity of source in 2-10 keV band.
    psp_table : object
        photospline.SplineTable object
    crit_log_nu_energy_lower : float
        Lower end of energy range (support) of spline flux.
    crit_log_nu_energy_upper : float
        Upper end of energy range (support) of spline flux.
    src_dist : float
        Distance to source in units of Mpc.
    lumin_scale : float
        A relative flux scaling factor. Can correct cases when the model
        calculation has a different normalization from what is desired.
    crit_log_energy_flux : float
        The spline is parameterized in log10(flux). This value determines
        when the flux should be considered 0.
    """
    def __init__(
            self, psp_table, log_xray_lumin, src_dist, Phi0,
            lumin_scale=1.0,
            crit_log_energy_flux=-50,
            crit_log_nu_energy_lower=2.0,
            crit_log_nu_energy_upper=7.0):

        super(SeyfertCoreCoronaFlux, self).__init__(
            Phi0, psp_table, crit_log_nu_energy_lower, crit_log_nu_energy_upper)

        self._lumin_scale = lumin_scale
        self._crit_log_energy_flux = crit_log_energy_flux
        self._src_dist = src_dist
        self._log_xray_lumin = log_xray_lumin

    @property
    def log_xray_lumin(self):
        """The log10 of the intrinsic source luminosity in 2-10keV x-ray band.
        """
        return self._log_xray_lumin
    @log_xray_lumin.setter
    def log_xray_lumin(self, v):
        v = float_cast(
            v, 'Property log_xray_lumin must be castable to type float!')
        self._log_xray_lumin = v

    @property
    def lumin_scale(self):
        """Relative factor for model flux normalization correction.
        """
        return self._lumin_scale
    @lumin_scale.setter
    def lumin_scale(self, v):
        v = float_cast(
            v, 'Property lumin_scale must be castable to type float!')
        self._lumin_scale = v

    @property
    def src_dist(self):
        """The distance to the source in units of Mpc.
        """
        return self._src_dist
    @src_dist.setter
    def src_dist(self, v):
        v = float_cast(
            v, 'Property src_dist must be castable to type float!')
        self._src_dist = v

    @property
    def crit_log_energy_flux(self):
        """Defines critical log energy when the flux is considered to be 0.
        """
        return self._crit_log_energy_flux
    @crit_log_energy_flux.setter
    def crit_log_energy_flux(self, v):
        v = float_cast(
            v, 'Property crit_log_energy_flux must be castable to type float!')
        self._crit_log_energy_flux = v

    @property
    def math_function_str(self):
        return (
            f'dN/dE = {self.Phi0:.2f} * {self.lumin_scale:.2f} '
            f'* 10^(log10(f(E)) - 2*log10(E) - 2*log10({self.src_dist:.2f}), '
            f'with log_xray_lumin={self.log_xray_lumin:.2f}'
        )

    def __call__(self, E):
        """The flux value dN/dE at energy E.

        Parameters
        ----------
        E : float | 1D ndarray of float
            Evaluation energy [GeV]

        Returns
        -------
        flux : float | 1D ndarray of float
            Flux at energy E in units of GeV^-1 cm^-2 s^-1.
        """

        log_enu = np.log10(E)
        log_energy_flux = self.psp_table.evaluate_simple([log_enu])

        # Convert energy flux to particle flux accounting for source distance.
        flux = 10**(log_energy_flux - 2.0*log_enu - 2.0*np.log10(self.src_dist))

        # Have to take care of very small fluxes (set to 0 beyond critical
        # energy or below the critical flux).
        out_of_bounds1 = log_energy_flux < self.crit_log_energy_flux
        out_of_bounds2 = np.logical_or(log_enu < self.crit_log_nu_energy_lower,
                                       log_enu > self.crit_log_nu_energy_upper)
        flux[np.logical_or(out_of_bounds1, out_of_bounds2)] = 0

        return self.Phi0 * self.lumin_scale * flux

    def __deepcopy__(self, memo):
        """The photospline.SplineTable objects are strictly immutable.
           Hence no copy should be required, ever!
        """
        return SeyfertCoreCoronaFlux(
            self.psp_table, self.log_xray_lumin, self.src_dist, self.Phi0,
            self.lumin_scale, self.crit_log_energy_flux,
            self.crit_log_nu_energy_lower, self.crit_log_nu_energy_upper
        )

    def __hash__(self):
        """We use hash in
        `skyllh.core.source_hypothesis.get_fluxmodel_to_source_mapping()` for
        mapping fluxes to KDE PDFs. Seyfert model KDEs only depend on the
        `log_xray_lumin` parameter.
        """
        hash_arg = (self.log_xray_lumin,)
        return hash(hash_arg)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and
            self.log_xray_lumin == other.log_xray_lumin
        )


class PowerLawFlux(NormedFluxModel):
    """Power law flux of the form

        dN/(dEdAdt) = Phi0 * (E / E0)^(-gamma)

    The unit of dN/(dEdAdt) is [energy]^-1 [length]^-2 [time]^-1.
    By default the unit is GeV^-1 cm^-2 s^-1.
    """
    def __init__(self, Phi0, E0, gamma):
        """Creates a new power law flux object.

        Parameters
        ----------
        Phi0 : float
            Flux value (dN/(dEdAdt)) at E0 in unit
            [energy]^-1 [length]^-2 [time]^-1.
            By default that is GeV^-1 cm^-2 s^-1.
        E0 : float
            Normalization energy.
        gamma : float
            Spectral index
        """
        super(PowerLawFlux, self).__init__(Phi0, E0)
        self.gamma = gamma

    @property
    def gamma(self):
        return self._gamma
    @gamma.setter
    def gamma(self, v):
        v = float_cast(v, 'Property gamma must be castable to type float!')
        self._gamma = v

    @property
    def math_function_str(self):
        return "dN/dE = %.2e * (E / %.2e %s)^-%.2f" \
            % (self.Phi0, self.E0, self.energy_unit, self.gamma)

    def __call__(self, E):
        """The flux value dN/dE at energy E.

        Parameters
        ----------
        E : float | 1d numpy.ndarray of float
            Evaluation energy [GeV]

        Returns
        -------
        flux : float | 1d ndarray of float
            Flux at energy E in unit [energy]^-1 [length]^-2 [time]^-1.
            By default in GeV^-1 cm^-2 s^-1.
        """
        flux = self.Phi0 * np.power(E / self.E0, -self.gamma)
        return flux

    def get_integral(self, E_min, E_max):
        """Returns the integral value of the flux between the given energy
        range.

        Parameters
        ----------
        E_min : float | 1d numpy ndarray of float
            The lower energy bound of the integration.
        E_max : float | 1d numpy ndarray of float
            The upper energy bound of the integration.

        Returns
        -------
        integral : float | 1d ndarray of float
            The integral value(s) of the given integral range(s).
        """
        gamma = self.gamma

        # Handle special case for gamma = 1.
        if(gamma == 1):
            integral = self.Phi0 * self.E0 * (
                np.log(np.abs(E_max)) - np.log(np.abs(E_min)))
            return integral

        integral = (self.Phi0 / ((1.-gamma)*np.power(self.E0, -gamma)) *
            (np.power(E_max, 1.-gamma) - np.power(E_min, 1.-gamma)))

        return integral

    def get_inv_normed_cdf(self, x, E_min, E_max):
        """Calculates the inverse cumulative distribution function value for
        each given value of x, which is a number between 0 and 1.

        Parameters
        ----------
        x : float | 1d numpy ndarray of float
            The argument value(s) of the inverse cumulative distribution
            function. Must be between 0 and 1.
        E_min : float
            The lower energy edge of the flux to be considered.
        E_max : float
            The upper energy edge of the flux to be considered.

        Returns
        -------
        inv_normed_cdf : float | 1d numpy ndarray
            The energy value(s) from the inverse normed cumulative distribution
            function.
        """
        gamma = self.gamma

        if(gamma == 1):
            N_0 = np.log(E_max / E_min)
            inv_normed_cdf = E_min * np.exp(x * N_0)
            return inv_normed_cdf

        N_0 = E_max ** (1. - gamma) - E_min ** (1. - gamma)
        inv_normed_cdf = np.power(
            x * N_0 + E_min**(1. - gamma),
            (1. / (1. - gamma)))

        return inv_normed_cdf


class CutoffPowerLawFlux(PowerLawFlux):
    """Cut-off power law flux of the form

        dN/(dEdAdt) = Phi0 * (E / E0)^(-gamma) * exp(-E/Ecut)

    The unit of dN/(dEdAdt) is [energy]^-1 [length]^-2 [time]^-1.
    By default the unit is GeV^-1 cm^-2 s^-1.
    """
    def __init__(self, Phi0, E0, gamma, Ecut):
        """Creates a new cut-off power law flux object.

        Parameters
        ----------
        Phi0 : float
            Flux value (dN/(dEdAdt)) at E0 in unit
            [energy]^-1 [length]^-2 [time]^-1. By default the unit is
            GeV^-1 cm^-2 s^-1.
        E0 : float
            Normalization energy [GeV]
        gamma : float
            Spectral index
        Ecut : float
            Cut-off energy [GeV]
        """
        super(CutoffPowerLawFlux, self).__init__(Phi0, E0, gamma)
        self.Ecut = Ecut

    @property
    def Ecut(self):
        return self._Ecut
    @Ecut.setter
    def Ecut(self, val):
        val = float_cast(val, 'Property val must be castable to type float!')
        self._Ecut = val

    @property
    def math_function_str(self):
        return super(CutoffPowerLawFlux, self).math_function_str + ' * exp(-E / %.2e %s)'%(self.Ecut, self.energy_unit)

    def __call__(self, E):
        """The flux value dN/(dEdAdt) at energy E.

        Parameters
        ----------
        E : float | 1d numpy.ndarray of float
            Evaluation energy.

        Returns
        -------
        flux : float | 1d ndarray of float
            Flux at energy E in unit [energy]^-1 [length]^-2 [time]^-1.
            By default that is GeV^-1 cm^-2 s^-1.
        """
        flux = super(CutoffPowerLawFlux, self).__call__(E) * np.exp(-E / self.Ecut)
        return flux


class LogParabolaPowerLawFlux(NormedFluxModel):
    """Power law flux with an index which varies as a log parabola in energy of
    the form

        dN/(dEdAdt) = Phi0 * (E / E0)^(-(alpha + beta*log(E / E0)))

    The unit of dN/(dEdAdt) is [energy]^-1 [length]^-2 [time]^-1.
    By default the unit is GeV^-1 cm^-2 s^-1.
    """
    def __init__(self, Phi0, E0, alpha, beta):
        super(LogParabolaPowerLawFlux, self).__init__(Phi0, E0)
        self.alpha = alpha
        self.beta = beta

    @property
    def alpha(self):
        return self._alpha
    @alpha.setter
    def alpha(self, v):
        v = float_cast(v, 'Property alpha must be castable to type float!')
        self._alpha = v

    @property
    def beta(self):
        return self._beta
    @beta.setter
    def beta(self, v):
        v = float_cast(v, 'Property beta must be castable to type float!')
        self._beta = v

    @property
    def math_function_str(self):
        return 'dN/dE = %.2e * (E / %.2e %s)^(-(%.2e + %.2e * log(E / %.2e %s)))'%(self.Phi0, self.E0, self.energy_unit, self.alpha, self.beta, self.E0, self.energy_unit)

    def __call__(self, E):
        """The flux value dN/(dEdAdt) at energy E.

        Parameters
        ----------
        E : float | 1d numpy.ndarray of float
            The evaluation energy.

        Returns
        -------
        flux : float | 1d ndarray of float
            Flux at energy E in unit [energy]^-1 [length]^-2 [time]^-1.
            By default that is GeV^-1 cm^-2 s^-1.
        """
        flux = self.Phi0 * np.power(E / self.E0, -self.alpha - self.beta * np.log(E / self.E0))
        return flux



# -*- coding: utf-8 -*-
# Authors:
#   Dr. Martin Wolf <mail@martin-wolf.org>

import numpy as np

from skyllh.core.py import module_classname
from skyllh.core.debugging import get_logger
from skyllh.core.parameters import make_params_hash
from skyllh.core.pdf import PDF
from skyllh.core.pdfratio import SigSetOverBkgPDFRatio


class PDPDFRatio(SigSetOverBkgPDFRatio):
    def __init__(self, sig_pdf_set, bkg_pdf, cap_ratio=False, **kwargs):
        """Creates a PDFRatio instance for the public data.
        It takes a signal PDF set for different discrete gamma values.

        Parameters
        ----------
        sig_pdf_set : instance of PDSignalEnergyPDFSet
            The PDSignalEnergyPDFSet instance holding the set of signal energy
            PDFs.
        bkg_pdf : instance of PDDataBackgroundI3EnergyPDF
            The PDDataBackgroundI3EnergyPDF instance holding the background
            energy PDF.
        cap_ratio : bool
            Switch whether the S/B PDF ratio should get capped where no
            background is available. Default is False.
        """
        self._logger = get_logger(module_classname(self))

        super().__init__(
            pdf_type=PDF,
            signalpdfset=sig_pdf_set,
            backgroundpdf=bkg_pdf,
            **kwargs)

        # Construct the instance for the fit parameter interpolation method.
        self._interpolmethod_instance = self.interpolmethod(
            self._get_ratio_values, sig_pdf_set.fitparams_grid_set)

        self.cap_ratio = cap_ratio
        if self.cap_ratio:
            self._logger.info('The energy PDF ratio will be capped!')

            # Calculate the ratio value for the phase space where no background
            # is available. We will take the p_sig percentile of the signal
            # like phase space.
            ratio_perc = 99

            # Get the log10 reco energy values where the background pdf has
            # non-zero values.
            n_logE = bkg_pdf.get_binning('log_energy').nbins
            n_sinDec = bkg_pdf.get_binning('sin_dec').nbins
            bd = bkg_pdf._hist_logE_sinDec > 0
            log10_e_bc = bkg_pdf.get_binning('log_energy').bincenters
            self.ratio_fill_value_dict = dict()
            for sig_pdf_key in sig_pdf_set.pdf_keys:
                sigpdf = sig_pdf_set[sig_pdf_key]
                sigvals = sigpdf.get_pd_by_log10_reco_e(log10_e_bc)
                sigvals = np.broadcast_to(sigvals, (n_sinDec, n_logE)).T
                r = sigvals[bd] / bkg_pdf._hist_logE_sinDec[bd]
                # Remove possible inf values.
                r = r[np.invert(np.isinf(r))]
                val = np.percentile(r[r > 1.], ratio_perc)
                self.ratio_fill_value_dict[sig_pdf_key] = val
                self._logger.info(
                    f'The cap value for the energy PDF ratio key {sig_pdf_key} '
                    f'is {val}.')

        # Create cache variables for the last ratio value and gradients in
        # order to avoid the recalculation of the ratio value when the
        # ``get_gradient`` method is called (usually after the ``get_ratio``
        # method was called).
        self._cache_fitparams_hash = None
        self._cache_ratio = None
        self._cache_gradients = None

    @property
    def cap_ratio(self):
        """Boolean switch whether to cap the ratio where no background
        information is available (True) or use the smallest possible floating
        point number greater than zero as background pdf value (False).
        """
        return self._cap_ratio
    @cap_ratio.setter
    def cap_ratio(self, b):
        self._cap_ratio = b

    def _get_signal_fitparam_names(self):
        """This method must be re-implemented by the derived class and needs to
        return the list of signal fit parameter names, this PDF ratio is a
        function of. If it returns an empty list, the PDF ratio is independent
        of any signal fit parameters.

        Returns
        -------
        list of str
            The list of the signal fit parameter names, this PDF ratio is a
            function of. By default this method returns an empty list indicating
            that the PDF ratio depends on no signal parameter.
        """
        fitparam_names = self.signalpdfset.fitparams_grid_set.parameter_names
        return fitparam_names

    def _is_cached(self, tdm, fitparams_hash):
        """Checks if the ratio and gradients for the given set of fit parameters
        are already cached.
        """
        if((self._cache_fitparams_hash == fitparams_hash) and
           (len(self._cache_ratio) == tdm.n_selected_events)
          ):
            return True
        return False

    def _get_ratio_values(self, tdm, gridfitparams, eventdata):
        """Select the signal PDF for the given fit parameter grid point and
        evaluates the S/B ratio for all the given events.
        """
        sig_pdf_key = self.signalpdfset.make_pdf_key(gridfitparams)

        sig_prob = self.signalpdfset.get_pdf(sig_pdf_key).get_prob(tdm)
        if isinstance(sig_prob, tuple):
            (sig_prob, _) = sig_prob

        bkg_prob = self.backgroundpdf.get_prob(tdm)
        if isinstance(bkg_prob, tuple):
            (bkg_prob, _) = bkg_prob

        if len(sig_prob) != len(bkg_prob):
            raise ValueError(
                f'The number of signal ({len(sig_prob)}) and background '
                f'({len(bkg_prob)}) probability values is not equal!')

        m_nonzero_bkg = bkg_prob > 0
        m_zero_bkg = np.invert(m_nonzero_bkg)
        if np.any(m_zero_bkg):
            ev_idxs = np.where(m_zero_bkg)[0]
            self._logger.debug(
                f'For {len(ev_idxs)} events the background probability is '
                f'zero. The event indices of these events are: {ev_idxs}')

        ratio = np.empty((len(sig_prob),), dtype=np.double)
        ratio[m_nonzero_bkg] = sig_prob[m_nonzero_bkg] / bkg_prob[m_nonzero_bkg]

        if self._cap_ratio:
            ratio[m_zero_bkg] = self.ratio_fill_value_dict[sig_pdf_key]
        else:
            ratio[m_zero_bkg] = (sig_prob[m_zero_bkg] /
                                 np.finfo(np.double).resolution)

        # Check for positive inf values in the ratio and set the ratio to a
        # finite number. Here we choose the maximum value of float32 to keep
        # room for additional computational operations.
        m_inf = np.isposinf(ratio)
        ratio[m_inf] = np.finfo(np.float32).max

        return ratio

    def _calculate_ratio_and_gradients(self, tdm, fitparams, fitparams_hash):
        """Calculates the ratio values and ratio gradients for all the events
        given the fit parameters using the interpolation method for the fit
        parameter. It caches the results.
        """
        (ratio, gradients) =\
            self._interpolmethod_instance.get_value_and_gradients(
                tdm, eventdata=None, params=fitparams)

        # Cache the value and the gradients.
        self._cache_fitparams_hash = fitparams_hash
        self._cache_ratio = ratio
        self._cache_gradients = gradients

    def get_ratio(self, tdm, fitparams=None, tl=None):
        """Calculates the PDF ratio values for all the events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        fitparams : dict | None
            The dictionary with the parameter name-value pairs.
            It can be ``None``, if the PDF ratio does not depend on any
            parameters.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : (N_events,)-shaped 1d numpy ndarray of float
            The PDF ratio value for each trial event.
        """
        fitparams_hash = make_params_hash(fitparams)

        # Check if the ratio value is already cached.
        if self._is_cached(tdm, fitparams_hash):
            return self._cache_ratio

        self._calculate_ratio_and_gradients(tdm, fitparams, fitparams_hash)

        return self._cache_ratio

    def get_gradient(self, tdm, fitparams, fitparam_name):
        """Retrieves the PDF ratio gradient for the pidx'th fit parameter.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            the PDF ratio gradient values should get calculated.
        fitparams : dict
            The dictionary with the fit parameter values.
        fitparam_name : str
            The name of the fit parameter for which the gradient should get
            calculated.
        """
        fitparams_hash = make_params_hash(fitparams)

        # Convert the fit parameter name into the local fit parameter index.
        pidx = self.convert_signal_fitparam_name_into_index(fitparam_name)

        # Check if the gradients have been calculated already.
        if self._is_cached(tdm, fitparams_hash):
            return self._cache_gradients[pidx]

        # The gradients have not been calculated yet.
        self._calculate_ratio_and_gradients(tdm, fitparams, fitparams_hash)

        return self._cache_gradients[pidx]
